# cython: infer_types=True
# cython: cdivision=True
cimport cython
from os import path
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from cpython.exc cimport PyErr_CheckSignals
from libc.stdio cimport FILE, fopen, fclose, fread, fwrite, feof, fseek
from libc.errno cimport errno
from libc.string cimport memcpy
from libc.math cimport sqrt

from libc.stdlib cimport qsort
from libc.stdint cimport int32_t

from cymem.cymem cimport Pool
from preshed.maps cimport PreshMap, MapStruct, map_get

from .sparse cimport SparseArray

from ..extra.eg cimport Example
from ..extra.mb cimport Minibatch
from ..structs cimport SparseArrayC, SparseAverageC
from ..typedefs cimport class_t, count_t, time_t, feat_t
from ..linalg cimport Vec, VecVec
from .serialize cimport Writer
from .serialize cimport Reader


cdef class AveragedPerceptron:
    '''A linear model for online supervised classification.
    Expected use is via Cython --- the Python API is impoverished and inefficient.

    Emphasis is on efficiency for multi-class classification, where the number
    of classes is in the dozens or low hundreds.
    '''
    def __init__(self, templates, *args, **kwargs):
        self.time = 0
        self.weights = PreshMap()
        self.averages = PreshMap()
        self.lasso_ledger = PreshMap()
        self.mem = Pool()
        self.extracter = ConjunctionExtracter(templates)
        self.learn_rate = kwargs.get('learn_rate', 0.001)
        self.l1_penalty = kwargs.get('l1_penalty', 1e-8)
        self.momentum = kwargs.get('momentum', 0.9)

    def __dealloc__(self):
        cdef size_t feat_addr
        # Use 'raw' memory management, instead of cymem.Pool, for weights.
        # The memory overhead of cymem becomes significant here.
        if self.weights is not None:
            for feat_addr in self.weights.values():
                if feat_addr != 0:
                    PyMem_Free(<SparseArrayC*>feat_addr)
        if self.averages is not None:
            for feat_addr in self.averages.values():
                if feat_addr != 0:
                    feat = <SparseAverageC*>feat_addr
                    PyMem_Free(feat.avgs)
                    PyMem_Free(feat.times)

    def __call__(self, eg_or_mb):
        cdef Example eg
        cdef Minibatch mb
        if isinstance(eg_or_mb, Example):
            eg = eg_or_mb
            self.set_scoresC(eg.c.scores, eg.c.features, eg.c.nr_feat)
        else:
            mb = eg_or_mb
            for i in range(mb.c.i):
                self.set_scoresC(mb.c.scores(i), mb.c.features(i), mb.c.nr_feat(i))
        PyErr_CheckSignals()
        return eg_or_mb

    def update(self, Example eg):
        self(eg)
        self.updateC(&eg.c)
        return eg.loss

    def dump(self, loc):
        cdef Writer writer = Writer(loc, self.weights.length)
        cdef feat_t key
        cdef size_t feat_addr
        for i, (key, feat_addr) in enumerate(self.weights.items()):
            if feat_addr != 0:
                writer.write(key, <SparseArrayC*>feat_addr)
            if i % 1000 == 0:
                PyErr_CheckSignals()
        writer.close()

    def load(self, loc):
        cdef feat_t feat_id
        cdef SparseArrayC* feature
        cdef Reader reader = Reader(loc)
        self.weights = PreshMap(reader.nr_feat)
        cdef int i = 0
        while reader.read(self.mem, &feat_id, &feature):
            self.weights.set(feat_id, feature)
            if i % 1000 == 0:
                PyErr_CheckSignals()
            i += 1
    
    def apply_owed_L1(self):
        cdef size_t feat_addr
        cdef feat_t feat_id
        u = self.time * self.learn_rate * self.l1_penalty
        for feat_id, feat_addr in self.averages.items():
            if feat_addr != 0:
                feat = <SparseAverageC*>feat_addr
                update_averages(feat, self.time)
                l1_paid = <weight_t><size_t>self.lasso_ledger.get(feat_id)
                l1_paid += group_lasso(feat.curr, l1_paid, u)
                self.lasso_ledger.set(feat_id, <void*><size_t>l1_paid)
 
    def end_training(self):
        cdef size_t feat_addr
        self.apply_owed_L1()
        for feat_id, feat_addr in self.averages.items():
            if feat_addr != 0:
                feat = <SparseAverageC*>feat_addr
                update_averages(feat, self.time+1)
                W = feat.curr
                avg = feat.avgs
                while W.key >= 0:
                    if W.val != 0:
                        W.val = avg.val / (self.time+1)
                    W += 1
                    avg += 1

    def resume_training(self):		
        cdef feat_t feat_id		
        cdef size_t feat_addr		
        for i, (feat_id, feat_addr) in enumerate(self.weights.items()):		
            train_feat = <SparseAverageC*>self.averages.get(feat_id)		
            if train_feat == NULL:		
                if train_feat is NULL:		
                    msg = (feat_id)		
                    raise MemoryError(		
                        "Error allocating memory for feature: %s" % msg)
                weights = <const SparseArrayC*>feat_addr		
                train_feat.curr  = SparseArray.clone(weights)		
                train_feat.avgs = SparseArray.clone(weights)		
                train_feat.times = SparseArray.clone(weights)		
                self.averages.set(feat_id, train_feat)
   
    @property
    def L1(self):
        cdef long double l1 = 0.0
        cdef size_t feat_addr
        for feat_addr in self.weights.values():
            if feat_addr == 0: continue
            feat = <const SparseArrayC*>feat_addr
            i = 0
            while feat.key >= 0:
                if feat.val < 0 or feat.val > 0:
                    l1 += abs(feat.val)
                feat += 1
        return l1

    @property
    def nr_active_feat(self):
        n = 0
        cdef size_t feat_addr
        for feat_addr in self.weights.values():
            if feat_addr == 0:
                continue
            feat = <const SparseArrayC*>feat_addr
            while feat.key >= 0:
                if feat.val != 0:
                    n += 1
                    break
                feat += 1
        return n
 
    @property
    def nr_weight(self):
        n = 0
        cdef size_t feat_addr
        for feat_addr in self.weights.values():
            if feat_addr == 0:
                continue
            feat = <const SparseArrayC*>feat_addr
            while feat.key >= 0:
                if feat.val != 0:
                    n += 1
                feat += 1
        return n
    
    @property
    def nr_feat(self):
        return self.extracter.nr_templ

    cdef void set_scoresC(self, weight_t* scores,
            const FeatureC* feats, int nr_feat) nogil:
        # This is the main bottle-neck of spaCy --- where we spend all our time.
        # Typical sizes for the dependency parser model:
        # * weights_table: ~9 million entries
        # * n_feats: ~200
        # * scores: ~80 classes
        # 
        # I think the bottle-neck is actually reading the weights from main memory.
        cdef const MapStruct* weights_table = self.weights.c_map
        cdef int i
        cdef FeatureC feat
        for feat in feats[:nr_feat]:
            class_weights = <const SparseArrayC*>map_get(weights_table, feat.key)
            if class_weights != NULL:
                i = 0
                while class_weights[i].key >= 0:
                    scores[class_weights[i].key] += class_weights[i].val * feat.value
                    i += 1

    @cython.cdivision(True)
    cdef int updateC(self, const ExampleC* eg) except -1:
        self.time += 1
        guess = VecVec.arg_max_if_true(eg.scores, eg.is_valid, eg.nr_class)
        if eg.costs[guess] > 0:
            best = VecVec.arg_max_if_zero(eg.scores, eg.costs, eg.nr_class)
            for feat in eg.features[:eg.nr_feat]:
                self.update_weight(feat.key, best,  -feat.value * eg.costs[guess])
                self.update_weight(feat.key, guess, feat.value * eg.costs[guess])
    
    cpdef int update_weight(self, feat_t feat_id, class_t clas,
            weight_t grad) except -1:
        if grad == 0:
            return 0
        if len(self.averages) < len(self.weights):		
            self.resume_training()
        feat = <SparseAverageC*>self.averages.get(feat_id)
        if feat == NULL:
            feat = <SparseAverageC*>PyMem_Malloc(sizeof(SparseAverageC))
            if feat is NULL:
                msg = (feat_id, clas, grad)
                raise MemoryError("Error allocating memory for feature: %s" % msg)
            feat.curr  = SparseArray.init(clas, grad)
            feat.avgs  = SparseArray.init(clas, 0)
            feat.times = SparseArray.init(clas, <weight_t>self.time)
            self.averages.set(feat_id, feat)
            self.weights.set(feat_id, feat.curr)
        else:  
            i = SparseArray.find_key(feat.curr, clas)
            if i < 0:
                feat.curr = SparseArray.resize(feat.curr)
                feat.avgs = SparseArray.resize(feat.avgs)
                feat.times = SparseArray.resize(feat.times)
                self.weights.set(feat_id, feat.curr)
                i = SparseArray.find_key(feat.curr, clas)
            feat.curr[i].key = clas
            feat.avgs[i].key = clas
            # Apply the last round of updates, multiplied by the time unchanged
            feat.avgs[i].val -= (self.time - feat.times[i].val) * feat.curr[i].val
            feat.curr[i].val -= grad
            feat.times[i].val = self.time

    cpdef int update_weight_ftrl(
            self, feat_t feat_id, class_t clas, weight_t grad) except -1:
        if grad == 0:
            return 0
        feat = <SparseAverageC*>self.averages.get(feat_id)
        if feat == NULL:
            feat = <SparseAverageC*>PyMem_Malloc(sizeof(SparseAverageC))
            if feat is NULL:
                msg = (feat_id, clas, grad)
                raise MemoryError("Error allocating memory for feature: %s" % msg)
            feat.curr  = SparseArray.init(clas, 0)
            feat.mom1  = SparseArray.init(clas, 0)
            feat.mom2  = SparseArray.init(clas, 0)
            feat.avgs  = SparseArray.init(clas, 0)
            feat.times = SparseArray.init(clas, 0)
            self.averages.set(feat_id, feat)
            self.weights.set(feat_id, feat.curr)
            i = 0
        else:  
            i = SparseArray.find_key(feat.curr, clas)
            if i < 0:
                feat.curr = SparseArray.resize(feat.curr)
                feat.mom1 = SparseArray.resize(feat.mom1)
                feat.mom2 = SparseArray.resize(feat.mom2)
                feat.avgs = SparseArray.resize(feat.avgs)
                feat.times = SparseArray.resize(feat.times)
                self.weights.set(feat_id, feat.curr)
                i = SparseArray.find_key(feat.curr, clas)
            feat.curr[i].key = clas
            feat.mom1[i].key = clas
            feat.mom2[i].key = clas
            feat.avgs[i].key = clas
            feat.times[i].key = clas
            # Apply the last round of updates, multiplied by the time unchanged
            update_averages(feat, self.time)
        adam_update(&feat.curr[i].val, &feat.mom1[i].val, &feat.mom2[i].val,
            self.time, feat.times[i].val, grad, self.learn_rate, self.momentum) 
        feat.times[i].val = self.time
        # Apply cumulative L1 penalty, from here:
        # http://www.aclweb.org/anthology/P09-1054 
        l1_paid = <weight_t><size_t>self.lasso_ledger.get(feat_id)
        l1_total = self.time * self.learn_rate * self.l1_penalty
        l1_paid += group_lasso(feat.curr, l1_paid, l1_total)
        self.lasso_ledger.set(feat_id, <void*><size_t>l1_paid)


cdef void adam_update(weight_t* w, weight_t* m1, weight_t* m2,
        weight_t t, weight_t last_upd, weight_t grad, weight_t learn_rate, weight_t _) nogil:
     # Calculate update with Adam
     beta1 = 0.9
     beta2 = 0.999
     eps = 1e-08

     m1[0] = (m1[0] * beta1) + ((1-beta1) * grad)
     m2[0] = (m2[0] * beta2) + ((1-beta2) * grad**2)

     # Estimate the number of updates, using time from last update
     nr_update = (t * (last_upd / t)) + 1
     
     m1t = m1[0] / (1-beta1**nr_update)
     m2t = m2[0] / (1-beta2**nr_update)
    
     w[0] -= learn_rate * m1t / (sqrt(m2t) + eps)


cdef void update_averages(SparseAverageC* feat, weight_t time) nogil:
    W = feat.curr
    avg = feat.avgs
    times = feat.times
    while W.key >= 0:
        unchanged = time - times.val
        avg.val += unchanged * W.val
        times.val = time
        W += 1
        times += 1
        avg += 1


cdef weight_t group_lasso(SparseArrayC* weights, weight_t penalty_paid,
                          weight_t total_penalty) nogil:
    norm = 0.0
    i = 0
    while weights[i].key >= 0:
        if weights[i].val > 0:
            norm += weights[i].val
        else:
            norm -= weights[i].val
        i += 1
    # Find what we want the norm to be
    target = max(0, norm - (penalty_paid + total_penalty))
    while weights.key >= 0:
        # If weights[i].val is negative, we want to add anyway ---
        # so should all work out.
        # The ideea here is to reduce the norm of the feature
        # proportionally.
        weights.val = (weights.val/norm) * target
        weights += 1
    return target - norm



import numpy


def categorical_crossentropy(scores, labels):
    target = numpy.zeros(scores.shape)
    loss = 0.
    for i in range(len(labels)):
        target[i, int(labels[i])] = 1.
        loss += (1.0-scores[i, int(labels[i])])**2
    return scores - target, loss



cdef class SparseArray:
    def __init__(self, int clas, weight_t value):
        self.c = SparseArray.init(clas, value)

    def __dealloc__(self):
        PyMem_Free(self.c)

    def __getitem__(self, int key):
        cdef int i = SparseArray.find_key(self.c, key)
        if i >= 0:
            return self.c[i].val
        else:
            return 0

    def __setitem__(self, int key, weight_t value):
        cdef int i = SparseArray.find_key(self.c, key)
        if i < 0:
            self.c = SparseArray.resize(self.c)
            i = SparseArray.find_key(self.c, key)
        self.c[i] = SparseArrayC(key=key, val=value)

    def __iter__(self):
        cdef int i = 0
        while self.c[i].key >= 0:
            yield (self.c[i].key, self.c[i].val)
            i += 1



class ShapeError(ValueError):
    @classmethod
    def expected_batch(cls, locals_, globals_):
        return cls("Expected batch")
    
    @classmethod
    def dim_mismatch(cls, expected, observed):
        return cls("Dimension mismatch: %s vs %s" % (expected, observed))



from __future__ import unicode_literals, print_function

from .optimizers import Eve, Adam, linear_decay
from .util import minibatch, score_model

import random
import tqdm


class Trainer(object):
    def __init__(self, model, train_data, L2=0.0):
        self.ops = model.ops
        self.model = model
        self.optimizer = Eve(model.ops, 0.001)
        self.batch_size = 128
        self.nb_epoch = 1
        self.i = 0
        self.L2 = 0.0
        self.dropout = 0.9
        self.dropout_decay = 1e-4
        self._loss = 0.

    def __enter__(self):
        return self, self.optimizer

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.model.use_params(self.optimizer.averages)

    def __call__(self, data, gradient):
        return self.optimizer(data, gradient, L2=self.L2)

    def iterate(self, model, train_data, dev_X, dev_Y, nb_epoch=None):
        if nb_epoch is None:
            nb_epoch = self.nb_epoch
        orig_dropout = self.dropout
        for i in range(nb_epoch):
            random.shuffle(train_data)
            for batch in tqdm.tqdm(minibatch(train_data,
                                   batch_size=self.batch_size)):
                X, y = zip(*batch)
                yield X, y
                self.dropout = linear_decay(orig_dropout, self.dropout_decay,
                                            self.optimizer.nr_iter)
            acc = score_model(model, dev_X, dev_Y)
            stats = (acc, self._loss, self.dropout)
            print('Dev.: %.3f, loss %.3f. Drop %.2f' % stats)
            self._loss = 0.



from __future__ import print_function, unicode_literals
from .ops import NumpyOps, CupyOps
import time                                                

def timeit(method):

    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        model, X, y = args
        n_y = len(y)
        print('%r %d examples %2.2f sec' % (model.name, n_y, te-ts))
        return result

    return timed


def get_ops(ops):
    if ops in ('numpy', 'cpu'):
        return NumpyOps()
    elif ops == ('cupy', 'gpu'):
        return CupyOps()
    else:
        return ops


@timeit
def score_model(model, X, y):
    correct = 0
    total = 0
    scores = model.predict_batch(X)
    if isinstance(y, tuple) and (isinstance(y[0], tuple) or isinstance(y[0], list)):
        y = model.ops.asarray(model.ops.flatten(y), dtype='i')
    else:
        y = model.ops.asarray(y, dtype='i')
    for i, gold in enumerate(y):
        correct += scores[i].argmax() == gold
        total += 1
    return float(correct) / total


def partition(examples, split_size):
    examples = list(examples)
    numpy.random.shuffle(examples)
    n_docs = len(examples)
    split = int(n_docs * split_size)
    return examples[:split], examples[split:]


def minibatch(stream, batch_size=1000):
    batch = []
    for X in stream:
        batch.append(X)
        if len(batch) >= batch_size:
            yield batch
            batch = []
    if len(batch) != 0:
        yield batch


def im2col(sequence, window):
    assert window == 2
    output = numpy.zeros((sequence.shape[0], window*2+1, sequence.shape[1]))
    output[:, 2] = sequence
    output[2:, 0] = sequence[:-2]
    output[1:, 1] = sequence[:-1]
    output[:-1, 3] = sequence[1:]
    output[:-2, 4] = sequence[2:]
    # Words 0 and 1 have no LL feature
    assert sum(abs(output[0, 0])) == 0
    if len(output) >= 2:
        assert sum(abs(output[1, 0])) == 0
        # Word 0 no L feature
        assert sum(abs(output[1, 0])) == 0
    # Words -1 and -2 have no RR feature
    assert sum(abs(output[-1, 4])) == 0
    if len(output) >= 2:
        assert sum(abs(output[-2, 4])) == 0
        # Word -1 has no R feature
    assert sum(abs(output[-1, 3])) == 0
    return output


class Unassigned(object):
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __get__(self, obj, objtype):
        return None

    def __set__(self, obj, value):
        if not isinstance(obj, self.expected_type):
            raise TypeError("TODO Error")



from .model import Model


class Embed(Model):
    name = 'embed'
    
    @property
    def describe_params(self):
        init = self.ops.xavier_uniform_init
        yield 'W-%s' % self.name, (self.nr_out, self.vector_length), init

    @property
    def shape(self):
        if self.output_shape is None or self.input_shape is None:
            return None
        else:
            return (self.nr_out, self.vector_length)

    @property
    def output_shape(self):
        return (self.nr_out,) if self.nr_out is not None else None

    @property
    def input_shape(self):
        return (self.vector_length,) if self.vector_length is not None else None

    @property
    def W(self):
        return self.params.get('W-%s' % self.name, require=True)

    @property
    def d_W(self):
        return self.params.get('d_W-%s' % self.name, require=True)

    def __init__(self, nr_out, vector_length, vectors=None, **kwargs):
        self.nr_out = nr_out
        self.vector_length = vector_length
        self.vectors = vectors if vectors is not None else {}
        Model.__init__(self, **kwargs)

    def check_input(self, X):
        return True

    def add_vector(self, id_, vector_dim, add_gradient=True):
        if not hasattr(self, 'vectors') or self.vectors is None:
            self.vectors = {}
        param = self.ops.allocate(vector_dim)
        param[:] = self.ops.xp.random.uniform(-0.1, 0.1, vector_dim)
        self.vectors[id_] = param
        if add_gradient:
            if not hasattr(self, 'gradients'):
                self.gradients = {}
            self.gradients[id_] = self.ops.allocate(param.shape)

    def get_vector(self, id_):
        return self.vectors.get(id_)

    def get_gradient(self, id_):
        return self.gradients.get(id_)

    def predict_batch(self, ids):
        if len(ids) < 1000:
            vectors = self._embed(ids)
            return self.ops.batch_dot(vectors, self.W)
        id_map = {}
        for i, id_ in enumerate(ids):
            if id_ not in id_map:
                id_map[id_] = [i]
            else:
                id_map[id_].append(i)
        mapped = sorted(id_map.items())
        vectors = self._embed([id_ for id_, _ in mapped])
        result = self.ops.batch_dot(vectors, self.W)
        output = self.ops.allocate((len(ids), self.W.shape[0]))
        for i, (_, occurs) in enumerate(mapped):
            for j in occurs:
                output[j] = result[i]
        return output

    def begin_update(self, ids, dropout=0.0):
        for id_ in ids:
            vector = self.get_vector(id_)
            if vector is None:
                self.add_vector(id_, self.input_shape[0], add_gradient=True)
        return self.predict_batch(ids), self._get_finish_update(ids)

    def check_input(self, x, expect_batch=False):
        return True

    def _embed(self, ids):
        vectors = self.ops.allocate((len(ids), self.W.shape[1]))
        for i, id_ in enumerate(ids):
            vector = self.get_vector(id_)
            if vector is not None:
                vectors[i] = vector
        return vectors

    def _get_finish_update(self, ids):
        def finish_update(gradients, optimizer=None, **kwargs):
            d_W = self.d_W
            d_W += self.ops.batch_outer(gradients, self._embed(ids))
            gradients = self.ops.batch_dot(gradients, self.W.T)
            tuned = set()
            for id_, delta_in in zip(ids, gradients):
                embed_grad = self.gradients.get(id_)
                if embed_grad is not None:
                    embed_grad += delta_in
                    tuned.add(id_)
            if optimizer is not None:
                if not kwargs.get('is_child'):
                    optimizer(self.params.weights, self.params.gradient)
                for id_ in tuned:
                    vector = self.get_vector(id_)
                    grad = self.gradients.get(id_)
                    optimizer(vector, grad, key=(self.name, id_))
            return None
        return finish_update



#    nr_out = None
#    nr_in = None
#    data = None
#
#    @property
#    def is_initialized(self):
#        return self.W is not None
#
#    @property
#    def input_shape(self):
#        return (self.nr_in,)
#
#    @property
#    def output_shape(self):
#        return (self.nr_out,)
#
#    @property
#    def nr_weight(self):
#        return self.nr_out * self.nr_in
#
#    def setup(self, vectors, W, **kwargs):
#        self.vectors = vectors
#        self.W = W
#
#    def set_weights(self, data=None, initialize=True, example=None):
#        if example is not None:
#            self.nr_in = example.shape[-1]
#        if data is None:
#            if self.data is None:
#                self.data = self.ops.allocate_pool(self.nr_weight,
#                                name=(self.name, 'pool'))
#            data = self.data
#        self.W = data.allocate_shape((self.nr_out, self.nr_in))
#        if initialize:
#            self.ops.xavier_uniform_init(self.W, inplace=True)
#
#    def set_gradient(self, data=None, initialize=False):
#        if data is None:
#            self.d_data = self.ops.allocate_pool(self.nr_weight,
#                            name=(self.name, 'pool'))
#        else:
#            self.d_data = data
#        self.d_W = self.d_data.allocate_shape((self.nr_out, self.nr_in))
#        self.gradients = {}
#
#



import pytest
import numpy
from hypothesis import given
from numpy.testing import assert_allclose

from .. import strategies
from ...ops import NumpyOps


@pytest.fixture
def ops():
    return NumpyOps()


def test_get_dropout_empty(ops):
    shape = (2,2)
    drop = 0.0
    mask = ops.get_dropout_mask(shape, drop)
    if drop <= 0.0:
        assert mask is None
    else:
        assert mask is not None


def test_get_dropout_not_empty(ops):
    shape = (2,2)
    drop = 0.1
    mask = ops.get_dropout_mask(shape, drop)
    if drop <= 0.0:
        assert mask is None
    else:
        assert mask is not None
    assert mask.shape == shape
    assert all(value >= 0 for value in mask.flatten())


@pytest.mark.skip
@given(X=strategies.arrays_BI())
def test_dropout_forward(ops, X):
    drop_prob = 0.25
    def drop_first_cell(shape, drop_prob_):
        assert drop_prob_ == drop_prob
        drop_mask = numpy.ones(shape)
        drop_mask /= (1. - drop_prob)
        drop_mask[0, 0] = 0.
        return drop_mask

    ops.get_dropout_mask = drop_first_cell
    output, backprop = ops.dropout(X, drop_prob)
    assert output[0, 0] == 0.
    for i in range(1, output.shape[0]):
        for j in range(output.shape[1]):
            assert output[i, j] == X[i, j] * (1. / 0.75)

@pytest.mark.skip
@given(X=strategies.arrays_BI())
def test_dropout_backward(ops, X):
    drop_prob = 0.25
    def drop_first_cell(shape, drop_prob_):
        assert drop_prob_ == drop_prob
        drop_mask = numpy.ones(shape)
        drop_mask /= (1. - drop_prob)
        drop_mask[0, 0] = 0.
        return drop_mask

    ops.get_dropout_mask = drop_first_cell
    output, backprop = ops.dropout(X, drop_prob)
    gradient = numpy.ones(output.shape)
    def finish_update(d, *args, **kwargs):
        return d
    output_gradient = backprop(finish_update)(gradient)
    assert output_gradient[0, 0] == 0.
    for i in range(1, output.shape[0]):
        for j in range(output.shape[1]):
            assert output_gradient[i, j] == 1. * (4. / 3.)



@pytest.mark.skip
@given(X=strategies.arrays_BI())
def test_softmax_sums_to_one(ops, X):
    y = ops.softmax(X)
    for row in y:
        assert 0.99999 <= row.sum() <= 1.00001


@pytest.mark.skip
@given(X=strategies.arrays_BI())
def test_softmax_works_inplace(ops, X):
    ops.softmax(X, inplace=True)
    for row in X:
        assert 0.99999 <= row.sum() <= 1.00001


@pytest.mark.skip
@given(W_b_inputs=strategies.arrays_OI_O_BI())
def test_batch_dot_computes_correctly(ops, W_b_inputs):
    W, _, inputs = W_b_inputs
    y = ops.batch_dot(inputs, W)
    expected = numpy.tensordot(inputs, W, axes=[[1], [1]])
    assert_allclose(y, expected)


@pytest.mark.skip
@given(arrays_BI_BO=strategies.arrays_BI_BO())
def test_batch_outer_computes_correctly(ops, arrays_BI_BO):
    bi, bo = arrays_BI_BO
    assert bi.shape[0] == bo.shape[0]
    assert len(bi.shape) == len(bo.shape) == 2
    expected = numpy.tensordot(bo, bi, axes=[[0], [0]])
    assert expected.shape == (bo.shape[1], bi.shape[1])
    oi = ops.batch_outer(bo, bi)
    assert_allclose(oi, expected)


@pytest.mark.skip
@given(X=strategies.arrays_BI())
def test_norm_computes_correctly(ops, X):
    for row in X:
        assert_allclose([numpy.linalg.norm(row)], [ops.norm(row)],
            rtol=1e-04, atol=0.0001)


@pytest.mark.skip
@given(W_b_X=strategies.arrays_OI_O_BI())
def test_dot_computes_correctly(ops, W_b_X):
    W, b, X = W_b_X
    for x in X:
        expected = numpy.dot(W, x)
        y = numpy.dot(W, x)
        assert_allclose(expected, y)


@pytest.mark.skip
@given(X=strategies.arrays_BI())
def test_argmax_computes_correctly(ops, X):
    which = ops.argmax(X, axis=-1)
    for i in range(X.shape[0]):
        assert max(X[i]) == X[i, which[i]]


@pytest.mark.skip
@given(X=strategies.arrays_BI())
def test_clip_low_computes_correctly_for_zero(ops, X):
    expected = X * (X > 0.)
    y = ops.clip_low(X, 0.)
    assert_allclose(expected, y)


@pytest.mark.skip
@given(X=strategies.arrays_BOP())
def test_take_which_computes_correctly(ops, X):
    which = numpy.argmax(X, axis=-1)
    best = ops.take_which(X, which)
    for i in range(best.shape[0]):
        for j in range(best.shape[1]):
            assert best[i, j] == max(X[i, j])


@pytest.mark.skip
@given(X=strategies.arrays_BI())
def test_flatten_unflatten_roundtrip(ops, X):
    flat = ops.flatten([x for x in X])
    assert flat.ndim == 1
    unflat = ops.unflatten(flat, [len(x) for x in X])
    assert_allclose(X, unflat)



# cython: profile=True
# cython: cdivision=True
# cython: infer_types = True
cimport cython
from libc.string cimport memcpy, memset
from libc.math cimport exp, sqrt
from libc.stdlib cimport calloc, malloc, free

import numpy
from cytoolz import concat
from numpy import prod
from numpy cimport ndarray

from ..typedefs cimport weight_t


try:
    import cupy
except ImportError:
    cupy = None

try:
    import cytoolz as toolz
except ImportError:
    import toolz


class Ops(object):
    xp = None

    def __init__(self, xp=None):
        if xp is not None:
            self.xp = xp

    def dropout(self, x, dropout, inplace=False):
        if dropout <= 0.0:
            return x, lambda func: func
        mask = self.get_dropout_mask(x.shape, dropout)
        def wrap_backprop(backprop):
            def finish_update(gradient, *args, **kwargs):
                return backprop(gradient * mask, *args, **kwargs)
            return finish_update
        if inplace:
            x *= mask
            return x, wrap_backprop
        else:
            return x * mask, wrap_backprop

    def flatten(self, X):
        return self.asarray(list(concat(X)))
 
    def unflatten(self, X, lengths):
        unflat = []
        for length in lengths:
            unflat.append(X[:length])
            X = X[length:]
        assert len(X) == 0
        assert len(unflat) == len(lengths)
        return unflat

    def get_dropout_mask(self, shape, drop):
        if drop <= 0.0:
            return None
        elif drop >= 1.0:
            return self.allocate(shape)
        coinflips = self.xp.random.uniform(0., 1., shape)
        return (coinflips >= drop) / (1.-drop)

    def allocate(self, shape):
        if isinstance(shape, int):
            shape = (shape,)
        nr_weight = numpy.prod(shape)
        return self.xp.zeros(shape, dtype='f')

    def asarray(self, data, dtype='f'):
        return self.xp.asarray(data, dtype=dtype)

    def batch_dot(self, x, y):
        return self.xp.tensordot(x, y, axes=[[1], [1]])
   
    def batch_outer(self, x, y):
        return self.xp.tensordot(x, y, axes=[[0], [0]])

    def norm(self, x):
        return self.xp.sqrt((x * x).sum())

    def dot(self, x, y):
        return self.xp.dot(x, y)
    
    def affine(self, weights, bias, signal):
        return self.batch_dot(signal, weights) + bias

    def argmax(self, x, axis=-1):
        return self.xp.argmax(x, axis=axis)

    def softmax(self, x, inplace=False, axis=1):
        if x.ndim >= 3:
            raise NotImplementedError(
                "Softmax currently only supports 2d. ndim=%d" % x.ndim)
        shape = x.shape
        maxes = self.xp.max(x, axis=1)
        maxes = maxes.reshape((x.shape[0], 1))
        shifted = x - maxes
        new_x = self.xp.exp(shifted)
        new_x /= new_x.sum(axis=1).reshape((x.shape[0], 1))
        if inplace:
            x[:] = new_x
            return x
        else:
            return new_x

    def expand_dims(self, a, axis=-1):
        return self.xp.expand_dims(a, axis=axis)

    def clip_low(self, x, value, inplace=False):
        if inplace:
            return self.xp.maximum(x, value, out=x)
        else:
            return self.xp.maximum(x, value)

    def take_which(self, x, which, axis=-1):
        output = self.allocate(which.shape)
        for i in range(x.shape[axis]):
            output += x[:,:,i] * (which == i)
        return output

    def xavier_uniform_init(self, W, inplace=True):
        scale = self.xp.sqrt(6. / (W.shape[0] + W.shape[1]))
        if inplace:
            W[:] = self.xp.random.uniform(-scale, scale, W.shape)
            return W
        else:
            return self.xp.random.uniform(-scale, scale, W.shape)

    def he_normal_init(self, shape, fan_in):
        scale = self.xp.sqrt(2. / fan_in)
        return self.xp.random.normal(scale=scale, size=prod(shape)).reshape(shape)


class NumpyOps(Ops):
    xp = numpy

    def elu(self, ndarray X, inplace=True):
        cdef weight_t* data = <weight_t*>X.data
        cdef size_t size = X.size
        for i in range(size):
            if data[i] < 0:
                data[i] = exp(data[i])-1.

    def backprop_elu(self, ndarray delta_, ndarray signal_out_,
            inplace=True):
        # Backprop the ELU transformation
        # Note that this is over the function _output_, not the function
        # _input_!
        cdef size_t size = delta_.size
        cdef weight_t* delta = <weight_t*>delta_.data
        cdef const weight_t* signal_out = <const weight_t*>signal_out_.data
        for i in range(size):
            if signal_out[i] <= 0:
                delta[i] *= signal_out[i] + 1.

    def relu(self, ndarray X, inplace=True):
        cdef weight_t* data = <weight_t*>X.data
        cdef size_t size = X.size
        for i in range(size):
            if data[i] < 0:
                data[i] = 0.

    def backprop_relu(self, ndarray delta_, ndarray signal_out_, inplace=True):
        cdef size_t size = delta_.size
        cdef weight_t* delta = <weight_t*>delta_.data
        cdef const weight_t* signal_out = <const weight_t*>signal_out_.data
        for i in range(size):
            if signal_out[i] <= 0:
                delta[i] = 0.


class CupyOps(Ops):
    xp = cupy



from .model import Model


class ScaleShift(Model):
    name = 'bnscale'
    @property
    def G(self):
        return self.params.get('G-%s' % self.name)

    @property
    def b(self):
        return self.params.get('b-%s' % self.name)

    @property
    def d_G(self):
        return self.params.get('d_G-%s' % self.name, require=True)

    @property
    def d_b(self):
        return self.params.get('d_b-%s' % self.name, require=True)

    @property
    def describe_params(self):
        '''
        Yields (name, shape, initializer) triples describing the weights directly
        owned by the layer.
        '''
        def init(G, **kwargs):
            G += 1
        yield 'G-%s' % self.name, (self.nr_out,), init
        yield 'b-%s' % self.name, (self.nr_out,), None

    def begin_update(self, input_BI, dropout=0.0, **kwargs):
        def finish_update(gradient_BI, *args, **kwargs):
            d_b = self.d_b
            d_b += gradient_BI.sum(axis=0)
            d_G = self.d_G
            for i in range(gradient_BI.shape[0]):
                d_G += gradient_BI[i] * input_BI[i]
            return gradient_BI * self.G
        return input_BI * self.G + self.b, finish_update

    def __init__(self, nr_out, *args, **kwargs):
        self.nr_out = nr_out
        Model.__init__(self, *args, **kwargs)
 

class BatchNormalization(Model):
    def check_input(self, X, expect_batch=True):
        return True

    def predict_batch(self, X):
        N, mu, var = _get_moments(self.ops, X)
        return _forward(self.ops, X, mu, var)
 
    def begin_update(self, X, dropout=0.0):
        N, mu, var = _get_moments(self.ops, X)
        Xhat = _forward(self.ops, X, mu, var)

        def finish_update(dy, optimizer=None, **kwargs):
            assert len(X) == len(dy)
            dist, sum_dy, sum_dy_dist = _get_d_moments(self.ops, dy, X, mu)
            if hasattr(dy, 'shape'):
                d_xhat = N * dy - sum_dy - dist * var**(-1.) * sum_dy_dist
                d_xhat *= var ** (-1. / 2)
                d_xhat /= N
                return d_xhat
            else:
                seqs = (dy, sum_dy, dist, sum_dy_dist)
                output = []
                assert len(sum_dy) == len(dy)
                assert len(dist) == len(dy)
                assert len(sum_dy_dist) == len(dy)
                for dy_, sum_dy_, dist_, sum_dy_dist_ in zip(*seqs):
                    d_xhat = N * dy_ - sum_dy_ - dist_ * var**(-1.) * sum_dy_dist_
                    d_xhat *= var ** (-1. / 2)
                    d_xhat /= N
                    output.append(d_xhat)
                assert len(output) == len(dy), (len(output), len(dy))
                return output
        return Xhat, finish_update


def _get_moments(ops, X):
    if hasattr(X, 'shape') and len(X.shape) == 2:
        mu = X.mean(axis=0)
        var = X.var(axis=0) + 1e-8
        return X.shape[0], mu, var
    else:
        stacked = numpy.vstack(X)
        return stacked.shape[0], stacked.mean(axis=0), stacked.var(axis=0)


def _get_d_moments(ops, dy, X, mu):
    if hasattr(dy, 'shape'):
        dist = X-mu
        return dist, ops.xp.sum(dy, axis=0), ops.xp.sum(dy * dist, axis=0)
    else:
        sum_dy = [ops.xp.sum(seq, axis=0) for seq in dy]
        dist = [x-mu for x in X]
        sum_dy_dot_dist = [ops.xp.sum(seq * d, axis=0) for seq, d in zip(dy, dist)]
        return dist, sum_dy, sum_dy_dot_dist


def _forward(ops, X, mu, var):
    if hasattr(X, 'shape'):
        return (X-mu) * var ** (-1./2.)
    else:
        return [_forward(x, mu, var) for x in X]



from __future__ import division

import pytest
import pickle
import io
import tempfile

from ..avgtron import AveragedPerceptron
from ...extra.eg import Example


def assert_near_eq(float1, float2):
    assert abs(float1 - float2) < 0.001


def test_basic():
    nr_class = 3
    model = AveragedPerceptron(((1,), (2,), (3,), (4,), (5,)))
    instances = [
        (1, {1: 1, 3: -5}),
        (2, {2: 4, 3: 5})
    ]
    for clas, feats in instances:
        eg = Example(nr_class)
        #eg.features = feats
        #model(eg)
        #eg.costs = [i != clas for i in range(nr_class)]
        #model.update(eg)
    #eg = Example(nr_class)
    #eg.features = {1: 2, 2: 1}
    #model(eg)
    #assert eg.guess == 2
    #eg = Example(nr_class)
    #eg.features = {0: 2, 2: 1}
    #model(eg)
    #assert eg.scores[1] == 0
    #eg = Example(nr_class)
    #eg.features = {1: 2, 2: 1}
    #model(eg)
    #assert eg.scores[2] > 0
    #eg = Example(nr_class)
    #eg.features = {1: 2, 1: 1}
    #model(eg)
    #assert eg.scores[1] > 0
    #eg = Example(nr_class)
    #eg.features = {0: 3, 3: 1}
    #model(eg)
    #assert eg.scores[1] < 0 
    #eg = Example(nr_class)
    #eg.features = {0: 3, 3: 1}
    #model(eg)
    #assert eg.scores[2] > 0 

#
#@pytest.fixture
#def instances():
#    instances = [
#        [
#            (1, {1: -1, 2: 1}),
#            (2, {1: 5, 2: -5}),
#            (3, {1: 3, 2: -3}),
#        ],
#        [
#            (1, {1: -1, 2: 1}),
#            (2, {1: -1, 2: 2}),
#            (3, {1: 3, 2: -3})
#        ],
#        [
#            (1, {1: -1, 2: 2}),
#            (2, {1: 5, 2: -5}), 
#            (3, {4: 1, 5: -7, 2: 1})
#        ]
#    ]
#    return instances
#
#@pytest.fixture
#def model(instances):
#    templates = []
#    for batch in instances:
#        for _, feats in batch:
#            for key in feats:
#                templates.append((key,))
#    templates = tuple(set(templates))
#    model = AveragedPerceptron(templates)
#    for batch in instances:
#        model.time += 1
#        for clas, feats in batch:
#            for key, value in feats.items():
#                model.update_weight(key, clas, value)
#    return model
#
#def get_score(nr_class, model, feats, clas):
#    eg = Example(nr_class)
#    eg.features = feats
#    eg.costs = [i != clas for i in range(nr_class)]
#    model(eg)
#    return eg.scores[clas]
#
#
#def get_scores(nr_class, model, feats):
#    eg = Example(nr_class)
#    eg.features = feats
#    model(eg)
#    return list(eg.scores)
#
#
#def test_averaging(model):
#    model.end_training()
#    nr_class = 4
#    # Feature 1
#    assert_near_eq(get_score(nr_class, model, {1: 1}, 1), sum([-1, -2, -3]) / 3.0)
#    assert_near_eq(get_score(nr_class, model, {1: 1}, 2), sum([5, 4, 9]) / 3.0)
#    assert_near_eq(get_score(nr_class, model, {1: 1}, 3), sum([3, 6, 6]) / 3.0)
#    # Feature 2
#    assert_near_eq(get_score(nr_class, model, {2: 1}, 1), sum([1, 2, 4]) / 3.0)
#    assert_near_eq(get_score(nr_class, model, {2: 1}, 2), sum([-5, -3, -8]) / 3.0)
#    assert_near_eq(get_score(nr_class, model, {2: 1}, 3), sum([-3, -6, -5]) / 3.0)
#    # Feature 3 (absent)
#    assert_near_eq(get_score(nr_class, model, {3: 1}, 1), 0)
#    assert_near_eq(get_score(nr_class, model, {3: 1}, 2), 0)
#    assert_near_eq(get_score(nr_class, model, {3: 1}, 3), 0)
#    # Feature 4
#    assert_near_eq(get_score(nr_class, model, {4: 1}, 1), sum([0, 0, 0]) / 3.0)
#    assert_near_eq(get_score(nr_class, model, {4: 1}, 2), sum([0, 0, 0]) / 3.0)
#    assert_near_eq(get_score(nr_class, model, {4: 1}, 3), sum([0, 0, 1]) / 3.0)
#    # Feature 5
#    assert_near_eq(get_score(nr_class, model, {5: 1}, 1), sum([0, 0, 0]) / 3.0)
#    assert_near_eq(get_score(nr_class, model, {5: 1}, 2), sum([0, 0, 0]) / 3.0)
#    assert_near_eq(get_score(nr_class, model, {5: 1}, 3), sum([0, 0, -7]) / 3.0)
#
#
#def test_dump_load(model):
#    loc = tempfile.mkstemp()[1]
#    model.end_training()
#    model.dump(loc)
#    string = open(loc, 'rb').read()
#    assert string
#    new_model = AveragedPerceptron([(1,), (2,), (3,), (4,)])
#    nr_class = 5
#    assert get_scores(nr_class, model, {1: 1, 3: 1, 4: 1}) != \
#           get_scores(nr_class, new_model, {1:1, 3:1, 4:1})
#    assert get_scores(nr_class, model, {2:1, 5:1}) != \
#            get_scores(nr_class, new_model, {2:1, 5:1})
#    assert get_scores(nr_class, model, {2:1, 3:1, 4:1}) != \
#           get_scores(nr_class, new_model, {2:1, 3:1, 4:1})
#    new_model.load(loc)
#    assert get_scores(nr_class, model, {1:1, 3:1, 4:1}) == \
#           get_scores(nr_class, new_model, {1:1, 3:1, 4:1})
#    assert get_scores(nr_class, model, {2:1, 5:1}) == \
#           get_scores(nr_class, new_model, {2:1, 5:1})
#    assert get_scores(nr_class, model, {2:1, 3:1, 4:1}) == \
#           get_scores(nr_class, new_model, {2:1, 3:1, 4:1})
#
#
### TODO: Need a test that exercises multiple lines. Example bug:
### in gather_weights, don't increment f_i per row, only per feature
### (so overwrite some lines we're gathering)



from collections import defaultdict

from .model import Model


class MaxoutWindowEncode(Model):
    name = 'encode'
    nr_piece = 3
    nr_feat = 5
    nr_out = None
    nr_in = None

    @property
    def describe_params(self):
        def _init(W, inplace=True):
            fan_in = W.shape[2] * W.shape[3]
            fan_out = W.shape[0]
            for i in range(W.shape[1]):
                scale = self.ops.xp.sqrt(2. / (fan_in + fan_out))
                W[:, i] = self.ops.xp.random.uniform(-scale, scale,
                                                     W[:, i].shape)
            return W
        yield 'W-%s' % self.name, self.shape, _init
        yield 'b-%s' % self.name, (self.nr_out, self.nr_piece), None

    @property
    def shape(self):
        if self.output_shape is None or self.input_shape is None:
            return None
        else:
            return (self.nr_out, self.nr_piece, self.nr_feat, self.nr_in)

    @property
    def output_shape(self):
        return (self.nr_out,) if self.nr_out is not None else None

    @property
    def input_shape(self):
        return (self.nr_feat, self.nr_in) if self.nr_in is not None else None

    @property
    def W(self):
        return self.params.get('W-%s' % self.name, require=True)
    
    @property
    def b(self):
        return self.params.get('b-%s' % self.name, require=True)

    @property
    def d_W(self):
        return self.params.get('d_W-%s' % self.name, require=True)
    
    @property
    def d_b(self):
        return self.params.get('d_b-%s' % self.name, require=True)

    def __init__(self, nr_out, embed, **kwargs):
        self.nr_out = nr_out
        self.embed = embed
        Model.__init__(self, **kwargs)

    def initialize_params(self, train_data=None, add_gradient=True):
        if train_data is not None and self.nr_in is None:
            self.nr_in = self._get_vector_dim_from_input(train_data)
        assert self.shape is not None, "TODO: Error"
        for name, shape, init in self.describe_params:
            if name not in self.params:
                self.params.add(name, shape)
                if init is not None:
                    init(self.params.get(name), inplace=True)
        self.embed.initialize_params(train_data, add_gradient=add_gradient)

    def predict_batch(self, seqs):
        positions, vectors, lengths = self.embed.predict_batch(seqs)
        out, _ = self._forward(positions, vectors, lengths)
        return out

    def begin_update(self, seqs, dropout=0.0):
        (positions, vectors, lengths), fine_tune = self.embed.begin_update(seqs, dropout=dropout)
        flat_out, whiches = self._forward(positions, vectors, lengths)
        flat_out, bp_dropout = self.ops.dropout(flat_out, dropout, inplace=True)
        finish_update = self._get_finish_update(
                            fine_tune, positions, vectors, lengths, whiches)
        return flat_out, bp_dropout(finish_update)

    def _forward(self, positions, vectors, lengths):
        cands = _dot_ids(self.ops, self.W, positions, vectors, lengths)
        cands += self.b
        which = self.ops.argmax(cands)
        best = self.ops.take_which(cands, which)
        return best, which

    def _get_finish_update(self, fine_tune, positions, vectors_UI, lengths, whiches_BO):
        B = whiches_BO.shape[0]
        I = vectors_UI.shape[1]
        O = self.nr_out
        F = self.nr_feat
        P = self.nr_piece
        def finish_update(gradients_BO, optimizer=None, **kwargs):
            gradients_BOP = self.ops.allocate((B, O, P))
            _get_full_gradients(gradients_BOP, gradients_BO, whiches_BO)
            d_b = self.d_b
            d_b += gradients_BOP.sum(axis=0)
            
            inputs_BFI = self.ops.allocate((B, F, I))
            _get_full_inputs(inputs_BFI, positions, vectors_UI, lengths)
            # Bop,Bfi->opfi
            d_W = self.d_W
            d_W += self.ops.batch_outer(gradients_BOP, inputs_BFI)
            if fine_tune is not None:
                W_OPFI = self.W
                gradients_BFI = self.ops.xp.einsum(
                                    'bop,opfi->bfi', gradients_BOP, W_OPFI)
                gradients_BI = gradients_BFI.sum(axis=1)
                fine_tune(gradients_BI, optimizer=optimizer, **kwargs)
            return None
        return finish_update


def _get_uniq_vectors(positions, vectors):
    # Need to set id back into sequence
    # Or, really, need to fix API of WindowEncode tagger. Maybe insert
    # shim operation?
    uniq_vectors = []
    remapped_positions = {}
    for i, (key, key_positions) in enumerate(positions.items()):
        seq_idx = key_positions[0]
        uniq_vectors.append(vectors[seq_idx])
        remapped_positions[i] = key_positions
    return remapped_positions, uniq_vectors


def _get_positions(ids):
    positions = defaultdict(list)
    for i, id_ in enumerate(ids):
        positions[id_].append(i)
    return positions


def _dot_ids(ops, W, positions, vectors, lengths):
    window_size = int((W.shape[2]-1) / 2)
    total_length = sum(lengths)
    # Shift the input, so that we don't have to special-case the starts and
    # ends. We'll shift back afterwards.
    out__bop = ops.allocate((total_length+(window_size*2), W.shape[0], W.shape[1]))
    H__ufop = _compute_hidden_layer(ops, W, vectors, lengths)
    
    for vec_idx, tok_idxs in positions.items():
        v__fop = H__ufop[vec_idx]
        for i in tok_idxs:
            out__bop[i : i+5] += v__fop
    # Shift the output, to correct for the 'padding' shift above.
    out__bop = out__bop[window_size : -window_size]
    return out__bop


def _compute_hidden_layer(ops, W__opfi, vectors__ui, lengths):
    H__uopf = ops.xp.tensordot(vectors__ui, W__opfi, axes=[[1], [3]])
    H__ufop = H__uopf.transpose((0, 3, 1, 2))
    return H__ufop


def _get_full_gradients(flat_gradients, gradients, whiches):
    for i in range(flat_gradients.shape[-1]):
        flat_gradients[:, :, i] += gradients * (whiches == i)
    return flat_gradients


def _get_full_inputs(writeto, positions, vectors, lengths):
    for vec_idx, tok_idxs in sorted(positions.items()):
        writeto[vec_idx, 2] = vectors[vec_idx]
    # Col 2 has w
    vectors = writeto[:, 2]
    # Col 0 has LL
    writeto[2:, 0] = vectors[:-2]
    # Col 1 has L
    writeto[1:, 1] = vectors[:-1]
    # Col 3 has R
    writeto[:-1, 3] = vectors[1:]
    # Col 4 has RR
    writeto[:-2, 4] = vectors[2:]
    ## Now use the lengths to zero LL, L, R and RR features as appropriate.
    #_zero_features_past_sequence_boundaries(writeto, lengths)
    return writeto


def _zero_features_past_sequence_boundaries(flat__bfop, lengths):
    # Now use the lengths to zero LL, L, R and RR features as appropriate.
    assert flat__bfop.shape[0] == sum(lengths), \
        (flat__bfop.shape, sum(lengths))
    i = 0
    for n in lengths:
        if n == 0:
            continue
        seq__nfop = flat__bfop[i : i+n]
        seq__nfop[0, 0] = 0
        seq__nfop[0, 1] = 0
        if len(seq__nfop) >= 2:
            seq__nfop[1, 0] = 0
        seq__nfop[-1, 3] = 0
        seq__nfop[-1, 4] = 0
        if len(seq__nfop) >= 2:
            seq__nfop[-2, 4] = 0
        i += n



"""
This is a batched LSTM forward and backward pass, by Andrej Karpathy
https://gist.github.com/karpathy/587454dc0146a6ae21fc
"""
from __future__ import unicode_literals, print_function
import numpy as np
import code


class LSTM(Model):
    '''
    Code by Andrej Karpathy, here:
    https://gist.github.com/karpathy/587454dc0146a6ae21fc
    '''
    fancy_forget_bias_init = 3.
    nr_in = None
    nr_out = None
    
    def setup(self, **kwargs): 
        """ 
        Initialize parameters of the LSTM (both weights and biases in one matrix) 
        One might way to have a positive fancy_forget_bias_init number
        (e.g. maybe even up to 5, in some papers)
        """
        pass

    def set_weights(self, data=None, initialize=True):
        # +1 for the biases, which will be the first row of WLSTM
        if initialize:
            self.ops.xavier_init(self.W)
        #self.weights = np.random.randn(input_size + hidden_size + 1,
        #                               4 * hidden_size)
        #self.weights /= np.sqrt(input_size + hidden_size)
        #self.weights[0,:] = 0 # initialize biases to zero
        if self.fancy_forget_bias_init != 0:
            # forget gates get little bit negative bias initially to encourage
            # them to be turned off remember that due to Xavier initialization
            # above, the raw output activations from gates before
            # nonlinearity are zero mean and on order of standard deviation ~1
            self.W[0,hidden_size:2*hidden_size] = self.fancy_forget_bias_init

    def predict_batch(self, sequences):
        X, is_not_end = pack_sequences(sequences)
        do_lstm = begin_lstm_forward(self.W, X.shape[1], X.shape[2])
        out = np.zeros((X.shape[0], X.shape[1], self.nr_out))
        for t in range(X.shape[0]):
            out[t] = do_lstm(X[t], is_not_end[t])
        return unpack_sequences(out, is_not_end)
  
    def get_fwd_cache(self, X, c0 = None, h0 = None):
        """
        X should be of shape (n,b,input_size), where n = length of sequence,
        b = batch size
        """
        n, b, input_size = X.shape
        d = self.weights.shape[1]/4 # hidden size
        for t in range(n):
        cache = {}
        cache['Hout'] = Hout
        cache['IFOGf'] = IFOGf
        cache['IFOG'] = IFOG
        cache['C'] = C
        cache['Ct'] = Ct
        cache['Hin'] = Hin
        cache['c0'] = c0
        cache['h0'] = h0
        # return C[t], as well so we can continue LSTM with prev state init if needed
        return Hout, C[t], Hout[t], cache

    def begin_update(self, sequences):
        X, is_not_end = pack_sequences(sequences)
        do_lstm = begin_lstm_backward(self.W, X.shape[1], X.shape[2])
        out = np.zeros((X.shape[0], X.shape[1], self.nr_out))
        callbacks = [None for _ in range(X.shape[0])]
        for t in range(X.shape[0]):
            out[t], callbacks[t] = do_lstm(X[t], is_not_end[t])
        callbacks.reverse()
        def finish_update(gradient, optimizer=None, **kwargs):
            for callback in callbacks:
                gradient = callback(gradient)
            return gradient
        return out, finish_update

def begin_LSTM_forward(weights, batch_size, input_size):
    d = weights.shape[1]/4 # hidden size
    prevc = np.zeros((batch_size, hidden_size))
    prevh = np.zeros((batch_size, hidden_size))
    # input, forget, output, gate (IFOG)
    IFOG = np.zeros((batch_size, hidden_size * 4)) 
    C = np.zeros((b, d)) # cell content
    def fwd_lstm_step(X_t, is_not_end):
        # Perform the LSTM forward pass with X as the input
        # hidden representation of the LSTM (gated cell content)
        # concat [x,h] as input to the LSTM
        # (with 1 prepended for bias)
        Hin = np.hstack((np.ones((batch_size, 1)), X_t, prevh))
        # compute all gate activations. dots: (most work is this line)
        np.dot(Hin, self.weights, out=IFOG)
        # non-linearities, computed in-place
        IFO = IFOG[:,:3*d]
        G = IFOG[:,3*d:]
        IFO *= -1
        np.exp(IFO, out=IFO)
        IFO += 1
        IFO **= -1
        np.tanh(G, out=G) # tanh
        I = IFO[:,:d]
        F = IFO[:,d:2*d]
        O = IFO[:,2*d:3*d]
        # compute the cell activation
        C *= F
        C += I * G
        Hout = O * np.tanh(C)
        prevh[:] = Hout
        # Reset states at boundaries.
        prevh *= is_not_end
        IFOG *= is_not_end
        C *= is_not_end
        return Hout
    return fwd_lstm_step


def begin_lstm_backward(weights, d_weights, batch_size, input_size):
    C = np.zeros((b,d))
    H = np.zeros((b,d))
    
    # Perform the LSTM forward pass with X as the input
    xphpb = weights.shape[0] # x plus h plus bias, lol
    Hin = np.zeros((n, b, xphpb)) # input [1, xt, ht-1] to each tick of the LSTM
    Hout = np.zeros((n, b, d)) # hidden representation of the LSTM (gated cell content)
    IFOG = np.zeros((n, b, d * 4)) # input, forget, output, gate (IFOG)
    IFOGf = np.zeros((n, b, d * 4)) # after nonlinearity
    C = np.zeros((n, b, d)) # cell content
    Ct = np.zeros((n, b, d)) # tanh of cell content

    # backprop the LSTM
    dIFOG = np.zeros(IFOG.shape)
    dIFOGf = np.zeros(IFOGf.shape)
    dWLSTM = np.zeros(self.weights.shape)
    dHin = np.zeros(Hin.shape)
    dC = np.zeros(C.shape)
    dX = np.zeros((n,b,input_size))
    dh0 = np.zeros((b, d))
    dc0 = np.zeros((b, d))
    dHout = dHout_in.copy() # make a copy so we don't have any funny side effects
    if dcn is not None:
        dC[n-1] += dcn.copy() # carry over gradients from later
    if dhn is not None:
        dHout[n-1] += dhn.copy()

    def bwd_lstm_step(X):
        # concat [x,h] as input to the LSTM
        prevh = Hout[t-1] if t > 0 else h0
        Hin[t,:,0] = 1 # bias
        Hin[t,:,1:input_size+1] = X[t]
        Hin[t,:,input_size+1:] = prevh
        # compute all gate activations. dots: (most work is this line)
        IFOG[t] = Hin[t].dot(self.weights)
        # non-linearities
        IFOGf[t,:,:3*d] = 1.0/(1.0+np.exp(-IFOG[t,:,:3*d])) # sigmoids; these are the gates
        IFOGf[t,:,3*d:] = np.tanh(IFOG[t,:,3*d:]) # tanh
        # compute the cell activation
        prevc = C[t-1] if t > 0 else c0
        C[t] = IFOGf[t,:,:d] * IFOGf[t,:,3*d:] + IFOGf[t,:,d:2*d] * prevc
        Ct[t] = np.tanh(C[t])
        Hout[t] = IFOGf[t,:,2*d:3*d] * Ct[t]

        return Hout[t], finish_update
    return lstm_step
    

def _get_finish_backward(dC, dprevC, dHout, dprevHout, dIFOGf, dWLSTM, dHin,
        weights, tanhCt, prevc, d):
    def finish_update(gradient, optimizer=None, **kwargs):
        dIFOGf[:,2*d:3*d] = tanhCt * dHout
        # backprop tanh non-linearity first then continue backprop
        dC += (1-tanhCt**2) * (IFOGf[:,2*d:3*d] * dHout)
        if t > 0:
            dIFOGf[:,:,d:2*d] = prevc * dC
            dprevc += IFOGf[:,d:2*d] * dC
        else:
            dIFOGf[:,d:2*d] = c0 * dC
            dc0 = IFOGf[:,d:2*d] * dC
        dIFOGf[:,:d] = IFOGf[:,3*d:] * dC
        dIFOGf[:,3*d:] = IFOGf[:,:d] * dC
            
        # backprop activation functions
        dIFOG[:,3*d:] = (1 - IFOGf[:,3*d:] ** 2) * dIFOGf[:,3*d:]
        y = IFOGf[:,:3*d]
        dIFOG[:,:3*d] = (y*(1.0-y)) * dIFOGf[:,:3*d]
 
        # backprop matrix multiply
        dWLSTM += np.dot(Hin.transpose(), dIFOG)
        dHin = dIFOG.dot(weights.transpose())
 
        # backprop the identity transforms into Hin
        dX = dHin[:,1:input_size+1]
        dprevHout += dHin[:,input_size+1:]
        # TODO


# -------------------
# TEST CASES
# -------------------


def checkSequentialMatchesBatch():
    """ check LSTM I/O forward/backward interactions """
    n,b,d = (5, 3, 4) # sequence length, batch size, hidden size
    input_size = 10
    LSTM = LSTMModel(input_size, d) # input size, hidden size
    X = np.random.randn(n,b,input_size)
    h0 = np.random.randn(b,d)
    c0 = np.random.randn(b,d)

    # sequential forward
    cprev = c0
    hprev = h0
    caches = [{} for t in xrange(n)]
    Hcat = np.zeros((n,b,d))
    for t in xrange(n):
        xt = X[t:t+1]
        _, cprev, hprev, cache = LSTM.get_fwd_cache(xt, cprev, hprev)
        caches[t] = cache
        Hcat[t] = hprev

    # sanity check: perform batch forward to check that we get the same thing
    H, _, _, batch_cache = LSTM.get_fwd_cache(X, c0, h0)
    assert np.allclose(H, Hcat), 'Sequential and Batch forward don''t match!'

    # eval loss
    wrand = np.random.randn(*Hcat.shape)
    loss = np.sum(Hcat * wrand)
    dH = wrand

    # get the batched version gradients
    BdX, BdWLSTM, Bdc0, Bdh0 = LSTM.backward(dH, batch_cache)

    # now perform sequential backward
    dX = np.zeros_like(X)
    dWLSTM = np.zeros_like(LSTM.weights)
    dc0 = np.zeros_like(c0)
    dh0 = np.zeros_like(h0)
    dcnext = None
    dhnext = None
    for t in reversed(xrange(n)):
        dht = dH[t].reshape(1, b, d)
        dx, dWLSTMt, dcprev, dhprev = LSTM.backward(dht, caches[t], dcnext, dhnext)
        dhnext = dhprev
        dcnext = dcprev

        dWLSTM += dWLSTMt # accumulate LSTM gradient
        dX[t] = dx[0]
        if t == 0:
            dc0 = dcprev
            dh0 = dhprev

    # and make sure the gradients match
    print('Making sure batched version agrees with sequential version: (should all be True)')
    print(np.allclose(BdX, dX))
    print(np.allclose(BdWLSTM, dWLSTM))
    print(np.allclose(Bdc0, dc0))
    print(np.allclose(Bdh0, dh0))
  

def checkBatchGradient():
    """ check that the batch gradient is correct """

    # lets gradient check this beast
    n,b,d = (5, 3, 4) # sequence length, batch size, hidden size
    input_size = 10
    LSTM = LSTMModel(input_size, d) # input size, hidden size
    X = np.random.randn(n,b,input_size)
    h0 = np.random.randn(b,d)
    c0 = np.random.randn(b,d)

    # batch forward backward
    H, Ct, Ht, cache = LSTM.get_fwd_cache(X, c0, h0)
    wrand = np.random.randn(*H.shape)
    loss = np.sum(H * wrand) # weighted sum is a nice hash to use I think
    dH = wrand
    dX, dWLSTM, dc0, dh0 = LSTM.backward(dH, cache)

    def fwd():
        h,_,_,_ = LSTM.forward(X, c0, h0)
        return np.sum(h * wrand)

    # now gradient check all
    delta = 1e-5
    rel_error_thr_warning = 1e-2
    rel_error_thr_error = 1
    tocheck = [X, LSTM.weights, c0, h0]
    grads_analytic = [dX, dWLSTM, dc0, dh0]
    names = ['X', 'WLSTM', 'c0', 'h0']
    for j in xrange(len(tocheck)):
        mat = tocheck[j]
        dmat = grads_analytic[j]
        name = names[j]
        # gradcheck
        for i in xrange(mat.size):
            old_val = mat.flat[i]
            mat.flat[i] = old_val + delta
            loss0 = fwd()
            mat.flat[i] = old_val - delta
            loss1 = fwd()
            mat.flat[i] = old_val

            grad_analytic = dmat.flat[i]
            grad_numerical = (loss0 - loss1) / (2 * delta)

            if grad_numerical == 0 and grad_analytic == 0:
                rel_error = 0 # both are zero, OK.
                status = 'OK'
            elif abs(grad_numerical) < 1e-7 and abs(grad_analytic) < 1e-7:
                rel_error = 0 # not enough precision to check this
                status = 'VAL SMALL WARNING'
            else:
                rel_error = abs(grad_analytic - grad_numerical) / abs(grad_numerical + grad_analytic)
                status = 'OK'
                if rel_error > rel_error_thr_warning: status = 'WARNING'
                if rel_error > rel_error_thr_error: status = '!!!!! NOTOK'

            # print stats
            print('%s checking param %s index %s (val = %+8f), analytic = %+8f, numerical = %+8f, relative error = %+8f' % (status, name, `np.unravel_index(i, mat.shape)`, old_val, grad_analytic, grad_numerical, rel_error))



if __name__ == "__main__":
    checkSequentialMatchesBatch()
    checkBatchGradient()
    print('every line should start with OK. Have a nice day!')


#
#
#def forward(W, x, U, h, i, a):
#    # Input and gate computation
#    a[t] = tanh( W[c] * x[t] + U[c] * h[t-1])
#    i[t] = sigma(W[i] * x[t] + U[i] * h[t-1])
#    f[t] = sigma(W[f] * x[t] + U[f] * h[t-1])
#    o[t] = sigma(W[o] * x[t] + U[o] * h[t-1])
#
#    # Memory cell update
#    c[t] = dot(i[t], a[t]) + dot(f[t], c[t-1])
#
#    # Compute output
#    h[t] = dot(o[t], tanh(c[t]))
#
#
#def backward(d_h, tanh_ct, o_t, a_t, i_t, c_t1, f_t):
#    d_o = dot(d_h, tanh_ct)
#    d_c += dot(d_h, o_t, 1-tanh2(c_t))
#    d_i = dot(d_c, a_t)
#    d_a = dot(d_c, i_t)
#    d_f = dot(d_c, c_t1)
#    d_c1 = dot(d_c, f_t)
#
#    d_zt = 



from numpy import prod
import contextlib

from .. import util
from ..train import Trainer
from ..exceptions import ShapeError
from ..ops import Ops
from ..params import Params


def is_batch(x):
    return True
    #return isinstance(x, list) or len(x.shape) >= 2


class Model(object):
    '''Model base class.'''
    name = 'model'
    device = 'cpu'
    Trainer = Trainer
    ops = None
    layers = []
    Params = Params

    @property
    def size(self):
        if self.shape is None:
            return None
        else:
            return prod(self.shape)

    @property
    def describe_params(self):
        for desc in []: # Need to be empty generator
            yield desc
    
    @property
    def input_shape(self):
        return self.layers[0].input_shape if self.layers else None
 
    @property
    def output_shape(self):
        return self.layers[0].input_shape if self.layers else None

    @property
    def shape(self):
        return self.output_shape + self.input_shape if self.layers else None

    def __init__(self, *layers, **kwargs):
        self.layers = []
        kwargs = self._update_defaults(**kwargs)
        if self.ops is None:
            self.ops = util.get_ops(self.device)
        self._root = kwargs.get('root', self)
        self._parent = kwargs.get('parent', None)
        if self._parent is not None:
            self.name = self._parent.name + '-' + self.name
        self.params = self.Params(self.ops)

        kwargs.setdefault('root', self)
        kwargs['parent'] = self
        for layer in layers:
            if isinstance(layer, Model):
                self.layers.append(layer)
            else:
                self.layers.append(layer(**kwargs))

    def _update_defaults(self, *args, **kwargs):
        new_kwargs = {}
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                new_kwargs[key] = value
        return new_kwargs
    
    def initialize_params(self, train_data=None, add_gradient=True):
        for name, shape, init in self.describe_params:
            if name not in self.params:
                self.params.add(name, shape)
                if init is not None:
                    init(self.params.get(name), inplace=True)
        for layer in self.layers:
            layer.params = self.params
            layer.initialize_params(train_data, add_gradient=add_gradient)

    def check_input(self, x, expect_batch=False):
        if self.layers:
            return self.layers[0].check_input(x, expect_batch=expect_batch)
        if is_batch(x):
            shape = x.shape[1:]
        else:
            raise ShapeError.expected_batch(locals(), globals())
        if shape != self.input_shape:
            raise ShapeError.dim_mismatch(self.input_shape, shape)
        else:
            return True

    def __call__(self, x):
        '''Predict a single x.'''
        self.check_input(x)
        if is_batch:
            return self.predict_batch(x)
        else:
            return self.predict_one(x)

    def predict_one(self, x):
        X = self.ops.expand_dims(x, axis=0)
        return self.predict_batch(X)[0]


    def predict_batch(self, X):
        y, _ = self.begin_update(X)
        return y

    def begin_training(self, train_data):
        self.initialize_params(train_data, add_gradient=True)
        return self.Trainer(self, train_data)
    
    def pipe(self, stream, batch_size=1000):
        for batch in util.minibatch(stream, batch_size):
            ys = self.predict_batch(batch)
            for y in ys:
                yield y

    def update(self, stream, batch_size=1000):
        for X, y in util.minibatch(stream, batch_size=batch_size):
            output, finish_update = self.begin_update(X)
            gradient = finish_update(y)
            yield gradient
    
    def begin_update(self, X, **kwargs):
        self.check_input(X, expect_batch=True)
        callbacks = []
        for layer in self.layers:
            X = self.ops.xp.ascontiguousarray(X, dtype='f')
            X, finish_update = layer.begin_update(X, **kwargs)
            callbacks.append(finish_update)
        return X, self._get_finish_update(callbacks)
    
    def _get_finish_update(self, callbacks):
        def finish_update(gradient, optimizer=None, **kwargs):
            for callback in reversed(callbacks):
                gradient = self.ops.xp.ascontiguousarray(gradient, dtype='f')
                gradient = callback(gradient, optimizer=optimizer,
                                is_child=True)
            if optimizer is not None and self.params is not None \
            and not kwargs.get('is_child'):
                optimizer(self.params.weights, self.params.gradient,
                    key=('', self.name), **kwargs)
            return gradient
        return finish_update

    @contextlib.contextmanager
    def use_params(self, params):
        if ('', self.name) in params:
            current = self.params.weights.copy()
            current[:] = self.params.weights
            self.params.weights[:] = params[('', self.name)]
            yield
            self.params.weights[:] = current



from .affine import Affine


class ReLu(Affine):
    name = 'relu'
    @property
    def describe_params(self):
        '''
        Yields (name, shape, initializer) triples describing the weights directly
        owned by the layer.
        '''
        #def init(W, **kwargs):
        #    W += self.ops.he_normal_init(W.shape, W.shape[-1])
        init = self.ops.xavier_uniform_init
        yield 'W-%s' % self.name, (self.nr_out, self.nr_in), init
        yield 'b-%s' % self.name, (self.nr_out,), None
    
    def predict_batch(self, X):
        output = Affine.predict_batch(self, X)
        return self.ops.clip_low(output, 0, inplace=True)

    def begin_update(self, input_BI, dropout=0.0):
        output_BO, finish_affine = Affine.begin_update(self, input_BI)
        output_copy = self.ops.xp.ascontiguousarray(output_BO, dtype='f')
        self.ops.relu(output_copy)
        def finish_update(gradient, *args, **kwargs):
            gradient = self.ops.xp.ascontiguousarray(gradient, dtype='f')
            self.ops.backprop_relu(gradient, output_copy)
            return finish_affine(gradient, *args, **kwargs)
        output_BO[:] = output_copy
        return output_BO, finish_update



from ..api import activation


@activation
def relu(x__BO, **kwargs):
    mask = x__BO > 0
    def finish_update(d__BO, *args, **kwargs):
        return d__BO * mask
    return x__BO * mask, finish_update


@activation
def softmax(x__BO, **kwargs):
    def finish_update(d__BO, *args, **kwargs):
        return d__BO
    xp = kwargs.get('xp')
    new_x = xp.zeros(shape=shape, dtype='f')
    for i in range(new_x.shape[0]):
        new_x[i] = xp.exp(x[i] - xp.max(x[i]))
        new_x[i] /= new_x[i].sum()
    if inplace:
        x[:] = new_x
        return x
    else:
        return new_x


@activation
def maxout(x__BOP, *args, **kwargs):
    which__BO = xp.argmax(x__BOP, axis=-1)
    x__BO = xp.zeros(which__BO.shape)
    for i in range(x__BOP.shape[-1]):
        x__BO += x__BOP[:,:,i] * (which__BO == i)
    shape = x__BOP.shape
    def finish_update(d__BO, *_, **_2):
        d__BOP = xp.zeros(shape)
        for i in range(shape[-1]):
            d__BOP[:, :, i] += d__BO * (which__BO == i)
        return d__BOP
    return x__BO, finish_update



from .affine import Affine


class Maxout(Affine):
    name = 'maxout'
    
    @property
    def describe_params(self):
        yield 'W', (self.nr_out, self.nr_in), self.ops.xavier_uniform_init
        yield 'b', (self.nr_out,), None

    def setup(self, nr_out=None, nr_in=None, nr_piece=None, **kwargs):
        self.nr_out = nr_out
        self.nr_in = nr_in
        self.nr_piece = nr_piece
        self.shape = (self.nr_out, self.nr_piece, self.nr_in)
        self.params = Params(self.ops)

    def predict_batch(self, input_bi):
        acts_bop = self.ops.xp.tensordot(input_bi, self.params.W,
                        axes=[[1], [-1]])
        acts_bop += self.params.b
        which_bo = self.ops.argmax(acts_bop, axis=-1)
        return self.ops.take_which(acts_bop, which_bo)

    def begin_update(self, input_BI, dropout=0.0):
        self.check_input(input_BI, expect_batch=True)
        W_OCI = self.params.W
        b_OC = self.params.b
        output_BOC = self.ops.xp.tensordot(input_BI, W_OCI, axes=[[1], [-1]])
        output_BOC += b_OC
        which_BO = self.ops.argmax(output_BOC, axis=-1)
        best_BO = self.ops.take_which(output_BOC, which_BO)
        best_BO, bp_dropout = self.ops.dropout(best_BO, dropout, inplace=True)
        finish_update = self._get_finish_update(input_BI, which_BO)
        return best_BO, bp_dropout(finish_update)

    def _get_finish_update(self, acts_BI, which_BO):
        def finish_update(d_acts_BO, optimizer=None, **kwargs):
            d_acts_BOP = self.ops.allocate((d_acts_BO.shape[0], self.nr_out,
                                           self.nr_piece))
            for i in range(self.nr_piece):
                d_acts_BOP[:, :, i] += d_acts_BO * (which_BO == i)
            self.d_b += d_acts_BOP.sum(axis=0)
            self.d_W += self.ops.xp.tensordot(d_acts_BOP, acts_BI,
                                              axes=[[0], [0]])
            # Bop,opi->Bi
            d_acts_BI = self.ops.xp.tensordot(d_acts_BOP, self.W,
                                              axes=[[1,2], [0, 1]])
            return d_acts_BI
        return finish_update



from .affine import Affine


class Softmax(Affine):
    name = 'softmax'
    
    @property
    def describe_params(self):
        yield 'W-%s' % self.name, (self.nr_out, self.nr_in), None
        yield 'b-%s' % self.name, (self.nr_out,), None
    
    def predict_batch(self, X):
        output = Affine.predict_batch(self, X)
        act = self.activate(output)
        return act

    def begin_update(self, X, dropout=0.0, **kwargs):
        return Affine.begin_update(self, X, dropout=0.0)

    def activate(self, X):
        return self.ops.softmax(X, axis=-1)





# cython: profile=True
# cython: cdivision=True
# cython: infer_types=True
cimport cython
from libc.string cimport memcpy, memset
from libc.math cimport exp, sqrt
from libc.stdlib cimport calloc, malloc, free

from collections import defaultdict
import numpy

from ..typedefs cimport weight_t
from ..linalg cimport Mat, MatMat, MatVec, VecVec, Vec, sqrt


def linear_decay(rate, decay, nr_upd):
    return rate * 1./(1. + decay * nr_upd)


@cython.cdivision(True)
cdef void _clip_gradient(weight_t* gradient, weight_t threshold, int nr_weight) nogil:
    # Clip gradient
    grad_norm = Vec.norm(gradient, nr_weight)
    if grad_norm >= threshold:
        Vec.mul_i(gradient, threshold / grad_norm, nr_weight)


@cython.cdivision(True)
cdef void _update_averages(weight_t* ema,
        const weight_t* weights, int nr_weight, weight_t t) nogil:
    cdef weight_t decay = (1.0 + t) / (10.0 + t)
    if decay > 0.9999:
        decay = 0.9999
    for i in range(nr_weight):
        ema[i] -= (1-decay) * (ema[i] - weights[i])


def update_averages(averages, key, weight_t[:] weights,
        nr_upd, max_decay=0.9999):
    if key not in averages:
        averages[key] = numpy.zeros((weights.size,), dtype='f')
    cdef weight_t[:] avg = averages[key]
    _update_averages(&avg[0],
        &weights[0], avg.shape[0], nr_upd)


def clip_gradient(weight_t[:] gradient, threshold):
    _clip_gradient(&gradient[0], threshold, gradient.shape[0])


class SGD(object):
    def __init__(self, ops, lr, momentum=0.0, decay=0.0, **settings):
        self.ops = ops
        self.alpha = lr
        self.mu = momentum
        self.decay = decay
        self.momentums = {}
        self.averages = {} if settings.get('averages', True) else None
        self.nr_update = defaultdict(int)

    def __call__(self, weights, gradient, key=None):
        self.nr_update[key] += 1
        nr_upd = self.nr_update[key]
        lr = self.lr(nr_upd)
        if key is None or self.mu == 0.0:
            weights -= lr * gradient
            gradient.fill(0)
        else:
            if key not in self.momentums:
                self.momentums[key] = numpy.zeros(weights.shape)
            momentum = self.momentums[key]
            momentum *= self.mu
            momentum += gradient * lr
            weights -= momentum
            gradient.fill(0)
        if self.averages is not None:
            update_averages(self.averages, key, weights, nr_upd)

    def lr(self, nr_upd):
        return linear_decay(self.alpha, self.decay, nr_upd)


class Adam(object):
    def __init__(self, ops, lr, beta1=0.90, beta2=0.999, eps=1e-08, decay=0.0):
        self.ops = ops
        self.mom1 = {}
        self.mom2 = {}
        self.averages = {}
        self.nr_update = defaultdict(int)
        self.last_seen = defaultdict(int)
        self.alpha = lr
        self.b1 = beta1
        self.b2 = beta2
        self.eps = eps
        self.decay = decay
        self.d = 1.
        self.f = 0.

    def lr(self, nr_upd):
        alpha = linear_decay(self.alpha, self.decay, nr_upd)
        fix1 = 1.- (self.b1 ** nr_upd)
        fix2 = 1.- (self.b2 ** nr_upd)
        return alpha * numpy.sqrt(fix2) / fix1

    @property
    def nr_iter(self):
        if not self.nr_update:
            return 0
        return max(self.nr_update.values())

    def __call__(self, weight_t[:] weights, weight_t[:] gradient, key=None):
        assert key is not None
        assert len(gradient) >= 1
        assert not self.ops.xp.isnan(weights).any()
        assert not self.ops.xp.isnan(gradient).any()
        if key not in self.mom1:
            self.mom1[key] = self.ops.allocate(weights.size)
        if key not in self.mom2:
            self.mom2[key] = self.ops.allocate(weights.size)
        self.nr_update[key] += 1
        nr_upd = self.nr_update[key]

        clip_gradient(gradient, len(gradient) / 100.)

        cdef weight_t[:] mom1 = self.mom1[key]
        cdef weight_t[:] mom2 = self.mom2[key]
        cdef weight_t lr = self.lr(nr_upd)
        cdef weight_t b1 = self.b1
        cdef weight_t b2 = self.b1
        cdef weight_t eps = self.eps
        
        _adam(&weights[0], &gradient[0], &mom1[0], &mom2[0],
            weights.shape[0], b1, b2, eps, lr)
        
        if self.averages is not None:
            update_averages(self.averages, key, weights, nr_upd)

    def set_loss(self, loss):
        pass


@cython.cdivision(True)
cdef void _adam(
    weight_t* weights, weight_t* gradient, weight_t* mom1, weight_t* mom2, 
        int nr_weight, weight_t beta1, weight_t beta2, weight_t eps,
        weight_t learn_rate) nogil:
    Vec.mul_i(mom1,
        beta1, nr_weight)
    VecVec.add_i(mom1,
        gradient, 1-beta1, nr_weight)
    for i in range(nr_weight):
        mom2[i] = (beta2 * mom2[i]) + ((1-beta2) * gradient[i] * gradient[i])
    # Here we assume this is calculated by the caller.
    #cdef weight_t a_t = learn_rate * sqrt(1-beta2**hp.t) / (1-beta1**hp.t)
    for i in range(nr_weight):
        weights[i] -= learn_rate * (mom1[i] / (sqrt(mom2[i]) + eps))
    memset(gradient, 0, sizeof(gradient[0]) * nr_weight)


class Eve(Adam):
    def __init__(self, *args, **kwargs):
        Adam.__init__(self, *args, **kwargs)
        self.b3 = 0.999
        self.lower_threshold = 0.1
        self.upper_threshold = 10
        self.d = 1.
        self.f = None

    def set_loss(self, loss):
        if self.f is None:
            self.f = loss
            return
        old_f = self.f
        d = self.d
        c = self._get_c(loss, old_f)
        new_f = c * loss
        r = abs(new_f - old_f) / min(new_f, old_f)
        new_d = d + (1 - self.b3) * (r - d)
        self.d = new_d
        self.f = new_f

    def _get_c(self, loss, old_f):
        if loss < old_f:
            delta = self.lower_threshold + 1.
            Delta = self.upper_threshold + 1.
        else:
            delta = 1. / (self.upper_threshold + 1.)
            Delta = 1. / (self.lower_threshold + 1.)
        return min(max(delta, loss / old_f), Delta)



from collections import defaultdict

from .base import Model


class WindowEncode(Model):
    nr_piece = 3
    nr_feat = 5
    vectors = None
    nr_out = None
    nr_in = None
    static = True

    @property
    def nr_weight(self):
        nr_W = self.nr_out * self.nr_in
        nr_b = self.nr_out
        return self.nr_feat * self.nr_piece * (nr_W + nr_b)

    def setup(self, *args, **kwargs):
        self.data = None
        self.W = None
        if kwargs.get('W') is not None:
            self.nr_out = kwargs.get('W').shape[0]
            self.nr_piece = kwargs.get('W').shape[1]
            self.nr_in = kwargs.get('W').shape[2]
        if self.nr_out is not None and self.nr_in is not None \
        and self.nr_piece is not None:
            self.set_weights(initialize=True)
            self.set_gradient()
        if kwargs.get('W') is not None:
            self.W[:] = kwargs.get('W')
        if kwargs.get('b') is not None:
            self.b[:] = kwargs.get('b')

    def set_weights(self, data=None, initialize=True, example=None):
        if data is None:
            if self.data is None:
                self.data = self.ops.allocate_pool(self.nr_weight,
                                name=(self.name, 'pool'))
            data = self.data
        self.W = data.allocate_shape((self.nr_out, self.nr_piece,
                                      self.nr_feat, self.nr_in))
        self.b = data.allocate_shape((self.nr_out, self.nr_piece))
        if initialize:
            self.ops.xavier_uniform_init(self.W, inplace=True)

    def set_gradient(self, data=None, initialize=False):
        if data is None:
            self.d_data = self.ops.allocate_pool(self.nr_weight,
                            name=(self.name, 'pool'))
        else:
            self.d_data = data
        self.d_W = self.d_data.allocate_shape(self.W.shape)
        self.d_b = self.d_data.allocate_shape(self.b.shape)

    def predict_batch(self, X):
        out, _ = self._forward(X)
        out = self.ops.flatten(out)
        return out

    def begin_update(self, ids, dropout=0.0):
        outputs, whiches = self._forward(ids)
        flat_out = self.ops.flatten(outputs)
        flat_out, bp_dropout = self.ops.dropout(flat_out, dropout,
                                               inplace=True)
        finish_update = self._get_finish_update(ids, flat_out, whiches)
        return flat_out, bp_dropout(finish_update)

    def add_vector(self, id_, shape, add_gradient=True):
        if not hasattr(self, 'vectors') or self.vectors is None:
            self.vectors = {}
        param = self.ops.allocate(shape)
        param[:] = self.ops.xp.random.uniform(-0.1, 0.1, shape)
        self.vectors[id_] = param
        if add_gradient:
            if not hasattr(self, 'gradients'):
                self.gradients = {}
            self.gradients[id_] = self.ops.allocate(param.shape)

    def get_vector(self, id_):
        return self.vectors.get(id_)

    def get_gradient(self, id_):
        return self.gradients.get(id_)

    def _forward(self, batch):
        positions = self._get_positions(batch)
        dotted = self._dot_ids(positions, [len(seq) for seq in batch])
        out = [self.ops.allocate((len(x), self.nr_out)) for x in batch]
        whiches = []
        for i, cands in enumerate(dotted):
            cands += self.b
            which = self.ops.argmax(cands)
            best = self.ops.take_which(cands, which)
            out[i][:] = best
            whiches.append(which)
        return out, whiches

    def _get_positions(self, batch):
        ids = defaultdict(list)
        for i, seq in enumerate(batch):
            for j, id_ in enumerate(seq):
                ids[id_].append((i, j))
        for id_ in ids:
            vector = self.vectors.get(id_)
            if vector is None:
                self.add_param(id_, (self.nr_in,))
        return ids

    def _get_finish_update(self, ids, flat_out, whiches):
        def finish_update(flat_gradients, optimizer=None, **kwargs):
            lengths = [len(w) for w in whiches]
            
            all_inputs = self._get_all_inputs(ids)
            flat_whiches = self.ops.flatten(whiches)
            all_gradients = self._get_all_gradients(flat_gradients,
                                                    flat_whiches)
            if all_inputs.shape[0] == 0 or all_gradients.shape[0] == 0:
                return None
            self.d_b += all_gradients.sum(axis=0)
            # Bop,Bfi->opfi
            self.d_W += self.ops.batch_outer(all_gradients, all_inputs)
            tuned_ids = self._fine_tune(self.W, ids, all_gradients)
            if optimizer is not None:
                for id_ in tuned_ids:
                    optimizer(self.vectors.get(id_), self.get_gradient(id_),
                              key=(self.name, id_))
            return None
        return finish_update

    def _dot_ids(self, ids, lengths):
        out = [self.ops.allocate((length, self.nr_out, self.nr_piece))
                             for length in lengths]
        for id_, egs in ids.items():
            vector = self.vectors.get(id_)
            if vector is None:
                continue
            # opFi,i->Fop
            hidden = self.ops.xp.tensordot(self.W, vector, axes=[[3], [0]])
            hidden = hidden.transpose((2, 0, 1))
            for i, j in egs:
                out_i = out[i]
                if j >= 2 and (j+2) < lengths[i]:
                    out_i[j-2:j+3] += hidden
                else:
                    if j >= 2:
                        out_i[j-2] += hidden[0]
                    if j >= 1:
                        out_i[j-1] += hidden[1]
                    out_i[j] += hidden[2]
                    if (j+1) < lengths[i]:
                        out_i[j+1] += hidden[3]
                    if (j+2) < lengths[i]:
                        out_i[j+2] += hidden[4]
        return out

    def _fine_tune(self, W, X, all_gradients):
        if self.static:
            return set()
        # Bop,opfi->Bfi
        tuning = self.ops.xp.tensordot(all_gradients, W, axes=[[1,2], [0, 1]])
        tuned_ids = set()
        i = 0
        for ids in X:
            for id_ in ids:
                d_vector = self.get_gradient(id_)
                if d_vector is None:
                    i += 1
                    continue
                if i >= 2:
                    d_vector += tuning[i-2, 0]
                if i >= 1:
                    d_vector += tuning[i-1, 1]
                d_vector += tuning[i, 2]
                if (i+1) < tuning.shape[0]:
                    d_vector += tuning[i+1, 3]
                if (i+2) < tuning.shape[0]:
                    d_vector += tuning[i+2, 4]
                i += 1
                tuned_ids.add(id_)
                assert not self.ops.xp.isnan(d_vector).any()
        return tuned_ids

    def _get_all_inputs(self, X):
        total_length = sum(len(x) for x in X)
        all_inputs = self.ops.allocate((total_length, 5, self.nr_in))
        i = 0
        for ids in X:
            vectors = [self.vectors.get(id_) for id_ in ids]
            for vector in vectors:
                if i >= 2:
                    all_inputs[i-2, 0] = vector
                if i >= 1:
                    all_inputs[i-1, 1] = vector
                all_inputs[i, 2] = vector
                if (i+1) < total_length:
                    all_inputs[i+1, 3] = vector
                if (i+2) < total_length:
                    all_inputs[i+2, 4] = vector
                i += 1
        return all_inputs

    def _get_all_gradients(self, gradients, flat_whiches):
        all_gradients = self.ops.allocate((len(gradients), self.nr_out,
                                           self.nr_piece))
        for i in range(self.nr_piece):
            all_gradients[:, :, i] += gradients * (flat_whiches == i)
        return all_gradients



import pytest
import numpy

from ..._classes.window_encode import MaxoutWindowEncode
from ...optimizers import SGD


class MyWindowEncode(MaxoutWindowEncode):
    device = 'cpu'

    def add_param(self, id_, shape):
        pass


@pytest.fixture
def model():
    model = MyWindowEncode(device='cpu', nr_in=10, nr_out=3, vectors={})
    assert model.nr_out == 3
    return model


@pytest.mark.skip
def test_init():
    model = MyWindowEncode(device='cpu', nr_out=10, nr_in=2, vectors={})
    assert model.nr_out == 10
    assert model.W.shape == (10, 3, 5, 2)


@pytest.mark.skip
def test_get_ids(model):
    x = [[1, 10], [20, 20]]
    ids = model._get_positions(x)
    assert ids == {1: [(0, 0)], 10: [(0, 1)], 20: [(1, 0), (1, 1)]}


@pytest.mark.skip
def test_dot_ids(model):
    x = [[1, 10], [20, 20]]
    ids = model._get_positions(x)
    assert ids == {1: [(0, 0)], 10: [(0, 1)], 20: [(1, 0), (1, 1)]}
    model._dot_ids(ids, [2, 2])


@pytest.mark.xfail
def test_predict_batch(model):
    x = [[1, 10], [20, 20]]
    scores = model.predict_batch(x)


@pytest.mark.xfail
def test_update(model):
    x = [[1, 10], [20, 20]]
    scores, update = model.begin_update(x)
    grads = [np.zeros((len(eg), model.nr_out)) for eg in x]
    update(grads)


@pytest.mark.xfail
def set_vector(model, key, values):
    model.add_vector(key, (len(values),))
    vector = model.get_vector(key)
    for i, value in enumerate(values):
        vector[i] = value
 

@pytest.mark.xfail
def test_predict_shapes():
    model = WindowEncode('encode', nr_out=5, nr_in=2)
    set_vector(model, 1, [1., -1.])
    set_vector(model, 2, [0.25, -0.25])

    output = model.predict_batch(numpy.asarray([[1], [2, 1]]))
    assert len(output) == 2
    assert output[0].shape == (1, 5)
    assert output[1].shape == (2, 5)


@pytest.mark.xfail
def test_update():
    model = WindowEncode(nr_out=2, nr_in=2, nr_piece=2)
    model.W.fill(2)
    model.b.fill(0) 
    set_vector(model, 2, [0.25, -0.25])
    
    output, finish_update = model.begin_update(numpy.asarray([[2]]))
    finish_update(numpy.asarray([[[0.25, 0.25]]]), SGD(model.ops, 1.0))
    vector = model.get_vector(2)
    # Not positive these are correct?
    assert vector[0] == -0.75
    assert vector[1] == -1.25
    assert list(model.b[0]) == [0.0, 0.0]



from .base import Model


class MeanPooling(Model):
    def predict_batch(self, X):
        means = []
        for x in X:
            means.append(self.ops.mean(x, axis=0))
        return self.ops.asarray(means)

    def begin_update(self, X, dropout=0.0):
        mask = self.ops.get_dropout(X.shape, drop)
        X *= mask

        def finish_update(gradient):
            batch_grads = []
            for i, x in enumerate(X):
                grad_i = self.ops.allocate(x.shape)
                if x.shape[0] != 0:
                    grad_i += gradient[i] / x.shape[0]
                grad_i *= mask[i]
                batch_grads.append(grad_i)
            return batch_grads
        return self.predict_batch(X), finish_update


class MaxPooling(Model):
    def predict_batch(self, X):
        means = []
        for x in X:
            means.append(self.ops.max(x, axis=0))
        return self.ops.asarray(means)

    def predict_batch(self, X):
        maxes = []
        for x in X:
            if x.shape[0] == 0:
                maxes.append(self.ops.allocate(x.shape[1:], dtype='f'))
            else:
                maxes.append(x.max(axis=0))
        return self.ops.asarray(maxes)

    def begin_update(self, X, dropout=0.0):
        mask = dropout(X, drop)
        def finish_update(gradient):
            mask = dropout(X, drop, inplace=True)
            batch_grads = []
            for i, x in enumerate(X):
                grad_i = self.ops.allocate(x.shape)
                if x.shape[0] != 0:
                    grad_i += gradient[i] * (x == x.max(axis=0))
                grad_i *= mask[i]
                batch_grads.append(grad_i)
            return batch_grads
        return self.predict_batch(X), finish_update


class MinPooling(Model):
    def predict_batch(self, X):
        maxes = []
        for x in X:
            if x.shape[0] == 0:
                maxes.append(self.ops.allocate(x.shape[1:]))
            else:
                maxes.append(x.min(axis=0))
        return self.ops.asarray(maxes)

    def begin_update(self, X, dropout=0.0):
        mask = dropout(X, drop, inplace=True)
        def finish_update(gradient, optimizer, L2=0.0):
            batch_grads = []
            for i, x in enumerate(X):
                grad_i = self.ops.allocate(x.shape)
                if x.shape[0] != 0:
                    grad_i += gradient[i] * (x == x.min(axis=0))
                grad_i *= mask[i]
                batch_grads.append(grad_i)
            return batch_grads
        return self.predict_batch(X), finish_update


class MultiPooling(NeuralNet):
    def predict_batch(self, X):
        return self.ops.concatenate([in_.predict_batch(X) for in_ in self.inputs], axis=1)
 
    def begin_update(self, X, dropout=0.0):
        output = []
        backward = []
        start = 0
        for input_ in self.inputs:
            out, finish = input_.begin_update(X, dropout=drop)
            output.append(out)
            end = start + input_.nr_out
            backward.append((finish, start, end))
            start = end
        return self.ops.concatenate(output, axis=1), self._get_finish_update(backward)

    def _get_finish_update(self, backward):
        def finish_update(gradient):
            assert len(self.inputs) == 3 # TODO
            seq_grads = [bwd(gradient[:, start : end])
                         for bwd, start, end in backward]
            summed = []
            for grads in zip(*seq_grads):
                summed.append(sum(grads)) 
            return summed
        return finish_update



from .model import Model


class ExtractWindow(Model):
    '''Add context to vectors in a sequence by concatenating n surrounding
    vectors.
    
    If the input is (10, 32) and n=1, the output will be (10, 96), with
    output[i] made up of (input[i-1], input[i], input[i+1]).
    '''
    n = 1
    @property
    def size(self):
        return 0

    def predict_batch(self, X):
        nr_feat = self.n * 2 + 1
        shape = (X.shape[0], nr_feat) + X.shape[1:]
        output = self.ops.allocate(shape)
        # Let L(w[i]) LL(w[i]), R(w[i]), RR(w[i])
        # denote the words one and two to the left and right of word i.
        # So L(w[i]) == w[i-1] for i >= 1 and {empty} otherwise.
        # output[i, n-1] will be L(w[i]) so w[i-1] or empty
        # output[i, n-2] will be LL(w[i]) so w[i-2] or empty
        # etc
        for i in range(self.n):
            output[i+1:, i] = X[:-(i+1)]
        # output[i, n] will be w[i]
        output[:, self.n] = X
        # Now R(w[i]), RR(w[i]) etc
        for i in range(1, self.n+1):
            output[:-i, self.n + i] = X[:-i]
        return output.reshape(shape[0], self.ops.xp.prod(shape[1:]))

    def begin_update(self, X, dropout=0.0):
        output = self.predict_batch(X)
        output, bp_dropout = self.ops.dropout(output, dropout)
        finish_update = self._get_finish_update()
        return output, bp_dropout(finish_update)

    def _get_finish_update(self):
        def finish_update(gradient, optimizer=None, **kwargs):
            nr_feat = self.n * 2 + 1
            shape = (gradient.shape[0], nr_feat, int(gradient.shape[-1] / nr_feat))
            gradient = gradient.reshape(shape)
            output = self.ops.allocate((gradient.shape[0], gradient.shape[-1]))
            # Word w[i+1] is the R feature of w[i]. So
            # grad(w[i+1]) += grad(R[i, n+1])
            # output[:-1] += gradient[1:, self.n+1] 
            for i in range(1, self.n+1): # As rightward features
                output[:-i] += gradient[i:, self.n+i]
            # Word w[i-1] is the L feature of w[i]. So
            # grad(w[i-1]) += grad(L[i, n-1])
            # output[1:] += gradient[:-1, self.n-1] 
            for i in range(1, self.n+1): # As leftward features 
                output[i:] += gradient[:-i, self.n-i]
            # Central column of gradient is the word's own feature
            output += gradient[:, self.n, :]
            return output
        return finish_update



# cython: profile=True
# cython: experimental_cpp_class_def=True
# cython: cdivision=True
# cython: infer_types=True

from __future__ cimport division
cimport cython
from libc.string cimport memset, memcpy
from libc.math cimport log, exp
import math

from cymem.cymem cimport Pool
from preshed.maps cimport PreshMap
from ..typedefs cimport hash_t


cdef class Beam:
    def __init__(self, class_t nr_class, class_t width, weight_t min_density=0.0):
        assert nr_class != 0
        assert width != 0
        self.nr_class = nr_class
        self.width = width
        self.min_density = min_density
        self.size = 1
        self.t = 0
        self.mem = Pool()
        self._parents = <_State*>self.mem.alloc(self.width, sizeof(_State))
        self._states = <_State*>self.mem.alloc(self.width, sizeof(_State))
        cdef int i
        self.histories = [[] for i in range(self.width)]
        self._parent_histories = [[] for i in range(self.width)]

        self.scores = <weight_t**>self.mem.alloc(self.width, sizeof(weight_t*))
        self.is_valid = <int**>self.mem.alloc(self.width, sizeof(weight_t*))
        self.costs = <weight_t**>self.mem.alloc(self.width, sizeof(weight_t*))
        for i in range(self.width):
            self.scores[i] = <weight_t*>self.mem.alloc(self.nr_class, sizeof(weight_t))
            self.is_valid[i] = <int*>self.mem.alloc(self.nr_class, sizeof(int))
            self.costs[i] = <weight_t*>self.mem.alloc(self.nr_class, sizeof(weight_t))

    def __len__(self):
        return self.size

    property score:
        def __get__(self):
            return self._states[0].score

    property min_score:
        def __get__(self):
            return self._states[self.size-1].score

    property loss:
        def __get__(self):
            return self._states[0].loss

    property probs:
        def __get__(self):
            return _softmax([self._states[i].score for i in range(self.size)])
 
    cdef int set_row(self, int i, const weight_t* scores, const int* is_valid,
                     const weight_t* costs) except -1:
        cdef int j
        for j in range(self.nr_class):
            self.scores[i][j] = scores[j]
            self.is_valid[i][j] = is_valid[j]
            self.costs[i][j] = costs[j]

    cdef int set_table(self, weight_t** scores, int** is_valid, weight_t** costs) except -1:
        cdef int i, j
        for i in range(self.width):
            memcpy(self.scores[i], scores[i], sizeof(weight_t) * self.nr_class)
            memcpy(self.is_valid[i], is_valid[i], sizeof(bint) * self.nr_class)
            memcpy(self.costs[i], costs[i], sizeof(int) * self.nr_class)
    
    cdef int initialize(self, init_func_t init_func, int n, void* extra_args) except -1:
        for i in range(self.width):
            self._states[i].content = init_func(self.mem, n, extra_args)
            self._parents[i].content = init_func(self.mem, n, extra_args)

    @cython.cdivision(True)
    cdef int advance(self, trans_func_t transition_func, hash_func_t hash_func,
                     void* extra_args) except -1:
        cdef weight_t** scores = self.scores
        cdef int** is_valid = self.is_valid
        cdef weight_t** costs = self.costs

        cdef Queue* q = new Queue()
        self._fill(q, scores, is_valid)
        # For a beam of width k, we only ever need 2k state objects. How?
        # Each transition takes a parent and a class and produces a new state.
        # So, we don't need the whole history --- just the parent. So at
        # each step, we take a parent, and apply one or more extensions to
        # it.
        self._parents, self._states = self._states, self._parents
        self._parent_histories, self.histories = self.histories, self._parent_histories
        cdef weight_t score
        cdef int p_i
        cdef int i = 0
        cdef class_t clas
        cdef _State* parent
        cdef _State* state
        cdef hash_t key
        cdef PreshMap seen_states = PreshMap(self.width)
        cdef uint64_t is_seen
        cdef uint64_t one = 1
        while i < self.width and not q.empty():
            data = q.top()
            p_i = data.second / self.nr_class
            clas = data.second % self.nr_class
            score = data.first
            q.pop()
            parent = &self._parents[p_i]
            # Indicates terminal state reached; i.e. state is done
            if parent.is_done:
                # Now parent will not be changed, so we don't have to copy.
                self._states[i] = parent[0]
                self._states[i].score = score
                self.histories[i] = list(self._parent_histories[p_i])
                i += 1
            else:
                state = &self._states[i]
                # The supplied transition function should adjust the destination
                # state to be the result of applying the class to the source state
                transition_func(state.content, parent.content, clas, extra_args)
                key = hash_func(state.content, extra_args) if hash_func is not NULL else 0
                is_seen = <uint64_t>seen_states.get(key)
                if key == 0 or key == 1 or not is_seen:
                    if key != 0 and key != 1:
                        seen_states.set(key, <void*>one)
                    state.score = score
                    state.loss = parent.loss + costs[p_i][clas]
                    self.histories[i] = list(self._parent_histories[p_i])
                    self.histories[i].append(clas)
                    i += 1
        del q
        self.size = i
        assert self.size >= 1
        for i in range(self.width):
            memset(self.scores[i], 0, sizeof(weight_t) * self.nr_class)
            memset(self.is_valid[i], 0, sizeof(int) * self.nr_class)
            memset(self.costs[i], 0, sizeof(weight_t) * self.nr_class)
        self.t += 1

    cdef int check_done(self, finish_func_t finish_func, void* extra_args) except -1:
        cdef int i
        self.is_done = True
        for i in range(self.size):
            if not self._states[i].is_done:
                self._states[i].is_done = finish_func(self._states[i].content, extra_args)
                if not self._states[i].is_done:
                    self.is_done = False

    @cython.cdivision(True)
    cdef int _fill(self, Queue* q, weight_t** scores, int** is_valid) except -1:
        """Populate the queue from a k * n matrix of scores, where k is the
        beam-width, and n is the number of classes.
        """
        cdef Entry entry
        cdef weight_t score
        cdef _State* s
        cdef int i, j, move_id
        assert self.size >= 1
        for i in range(self.size):
            s = &self._states[i]
            move_id = i * self.nr_class
            if s.is_done:
                # Update score by path average, following TACL '13 paper.
                if self.histories[i]:
                    entry.first = s.score + (s.score / self.t)
                else:
                    entry.first = s.score
                entry.second = move_id
                if q.empty() \
                or q.top().first < 0 \
                or entry.first >= (q.top().first * self.min_density):
                    q.push(entry)
            else:
                for j in range(self.nr_class):
                    if is_valid[i][j]:
                        entry.first = s.score + scores[i][j]
                        entry.second = move_id + j
                        if q.empty() \
                        or q.top().first < 0 \
                        or entry.first >= (q.top().first * self.min_density):
                            q.push(entry)


cdef class MaxViolation:
    def __init__(self):
        self.p_score = 0.0
        self.g_score = 0.0
        self.Z = 0.0
        self.gZ = 0.0
        self.delta = -1
        self.cost = 0
        self.p_hist = []
        self.g_hist = []
        self.p_probs = []
        self.g_probs = []

    cpdef int check(self, Beam pred, Beam gold) except -1:
        cdef _State* p = &pred._states[0]
        cdef _State* g = &gold._states[0]
        cdef weight_t d = p.score - g.score
        if p.loss >= 1 and (self.cost == 0 or d > self.delta):
            self.cost = p.loss
            self.delta = d
            self.p_hist = list(pred.histories[0])
            self.g_hist = list(gold.histories[0])
            self.p_score = p.score
            self.g_score = g.score
            self.Z = 1e-10
            self.gZ = 1e-10
            for i in range(pred.size):
                if pred._states[i].loss > 0:
                    self.Z += exp(pred._states[i].score)
            for i in range(gold.size):
                if gold._states[i].loss == 0:
                    prob = exp(gold._states[i].score)
                    self.Z += prob
                    self.gZ += prob

    cpdef int check_crf(self, Beam pred, Beam gold) except -1:
        d = pred.score - gold.score
        if pred.loss > 0 and (self.cost == 0 or d > self.delta):
            p_hist = []
            p_scores = []
            g_hist = []
            g_scores = []
            for i in range(pred.size):
                if pred._states[i].loss > 0:
                    p_scores.append(pred._states[i].score)
                    p_hist.append(list(pred.histories[i]))
                # This can happen from non-monotonic actions
                # If we find a better gold analysis this way, be sure to keep it.
                elif pred._states[i].loss == 0 and pred._states[i].score > gold.score:
                    g_scores.append(pred._states[i].score)
                    g_hist.append(list(pred.histories[i]))
            for i in range(gold.size):
                if gold._states[i].loss == 0:
                    g_scores.append(gold._states[i].score)
                    g_hist.append(list(gold.histories[i]))

            all_probs = _softmax(p_scores + g_scores)
            p_probs = all_probs[:len(p_scores)]
            g_probs_all = all_probs[len(p_scores):]
            g_probs = _softmax(g_scores)
            
            self.cost = pred.loss
            self.delta = d
            self.p_hist = p_hist
            self.g_hist = g_hist
            # TODO: These variables are misnamed! These are the gradients of the loss.
            self.p_probs = p_probs
            # Intuition here:
            # The gradient of the loss is:
            # P(model) - P(truth)
            # Normally, P(truth) is 1 for the gold
            # But, if we want to do the "partial credit" scheme, we want
            # to create a distribution over the gold, proportional to the scores
            # awarded.
            self.g_probs = [x-y for x, y in zip(g_probs_all, g_probs)]


def _softmax(nums):
    if not nums:
        return []
    max_ = max(nums)
    nums = [(exp(n-max_) if n is not None else None) for n in nums]
    Z = sum(n for n in nums if n is not None)
    return [(n/Z if n is not None else None) for n in nums]



import pytest


@pytest.mark.skip
def test_relu_clips_negative(model, input_BI):
    # TODO: There was a bug related to this
    pass


@pytest.mark.skip
def test_relu_clipped_gradients_are_zero():
    # TODO: There was a bug related to this
    pass



from .model import Model
from .batchnorm import BatchNormalization, ScaleShift
from ...api import layerize
from .affine import Affine

import cytoolz as toolz


class ResBlock(Model):
    name = 'resblock'

    @property
    def input_shape(self):
        return (self.nr_out,)

    @property
    def output_shape(self):
        return (self.nr_out,)

    def __init__(self, width, nr_in=None, **kwargs):
        self.nr_out = width
        self.nr_in = width if nr_in is None else nr_in
        self.name = kwargs.get('name')
        Model.__init__(self, **kwargs)
        self.layers = [
            BatchNormalization(name='%s-bn1' % self.name),
            ScaleShift(width, name='%s-scaleshift1' % self.name),
            layerize(_relu(self.ops)),
            Affine(width, width, name='%s-weight1' % self.name),
            #BatchNormalization(name='%s-bn2' % self.name),
            #ScaleShift(width, name='%s-scaleshift2' % self.name),
            #layerize(_relu(self.ops)),
            #Affine(width, width, name='%s-weight2' % self.name),
        ]


@toolz.curry
def _relu(ops, X, dropout=0.0):
    x_copy = ops.xp.ascontiguousarray(X, dtype='f')
    ops.relu(x_copy, inplace=True)
    mask = X > 0
    def finish_update(gradient, *args, **kwargs):
        gradient = ops.xp.ascontiguousarray(gradient, dtype='f')
        ops.backprop_relu(gradient, x_copy)
        return gradient
    X[:] = x_copy
    return X, finish_update


class Residual(Model):
    name = 'residual'

    Block = ResBlock

    @property
    def input_shape(self):
        return (self.nr_out,)

    @property
    def output_shape(self):
        return (self.nr_out,)

    def __init__(self, width, **kwargs):
        self.nr_out = width
        Model.__init__(self, **kwargs)
        if 'name' in kwargs:
            kwargs.pop('name')
        self.layers = [
            self.Block(width, width, name='rb1-%s' % self.name, **kwargs),
            self.Block(width, width, name='rb2-%s' % self.name, **kwargs)
        ]

    def begin_update(self, X, dropout=0.0):
        out1, upd1 = self.layers[0].begin_update(X, dropout=0.)
        out2, upd2 = self.layers[1].begin_update(out1, dropout=0.)
        output, bp_dropout = self.ops.dropout(out1 + out2, dropout)
        def finish_update(gradient, optimizer=None, **kwargs):
            grad2 = upd2(gradient, optimizer=optimizer, **kwargs)
            grad1 = upd1(grad2, optimizer=optimizer, **kwargs)
            return grad1 + grad2
        return output, bp_dropout(finish_update)



from .model import Model
from ..exceptions import ShapeError


class Affine(Model):
    '''Computes the linear transform Y = (W @ X) + b.

    See also: ReLu, Softmax, Maxout
    '''
    name = 'affine'
    
    @property
    def describe_params(self):
        '''
        Yields (name, shape, initializer) triples describing the weights directly
        owned by the layer.
        '''
        yield 'W-%s' % self.name, (self.nr_out, self.nr_in), self.ops.xavier_uniform_init
        yield 'b-%s' % self.name, (self.nr_out,), None

    @property
    def shape(self):
        if self.output_shape is None or self.input_shape is None:
            return None
        else:
            return (self.nr_out, self.nr_in)

    @property
    def output_shape(self):
        return (self.nr_out,) if self.nr_out is not None else None
    @output_shape.setter
    def output_shape(self, value):
        self.nr_out = value[0]

    @property
    def input_shape(self):
        return (self.nr_in,) if self.nr_in is not None else None
    @input_shape.setter
    def input_shape(self, value):
        self.nr_in = value[0]

    @property
    def W(self):
        return self.params.get('W-%s' % self.name, require=True)

    @property
    def b(self):
        return self.params.get('b-%s' % self.name, require=True)

    @property
    def d_W(self):
        return self.params.get('d_W-%s' % self.name, require=True)

    @property
    def d_b(self):
        return self.params.get('d_b-%s' % self.name, require=True)

    def __init__(self, nr_out=None, nr_in=None, *args, **kwargs):
        '''Construct an Affine object.

        Arguments:
            nr_out (int): Width of output vector.
            nr_in (int): Width of input vector.
        Keyword arguments:
            ops (Ops): Handler for mathematical operations. 
            device (str):
                Specify the device to execute the layer on,
                e.g. 'cpu', 'gpu0', 'gpu1'.
            Trainer (type): Class used by begin_training().
        '''
        self.nr_out = nr_out
        self.nr_in = nr_in
        # This sets attributes from kwargs.
        # args is passed for potential subclasses.
        Model.__init__(self, *args, **kwargs)

    def predict_batch(self, input_BI):
        return self.ops.affine(self.W, self.b, input_BI)

    def begin_update(self, input_BI, dropout=0.0):
        self.check_input(input_BI)
        output_BO = self.predict_batch(input_BI)
        if dropout != 0.0:
            output_BO, bp_dropout = self.ops.dropout(output_BO, dropout)
            return output_BO, bp_dropout(self._get_finish_update(input_BI))
        else:
            return output_BO, self._get_finish_update(input_BI)

    def _get_finish_update(self, acts_BI):
        def finish_update(d_acts_BO, optimizer=None, **kwargs):
            d_b = self.d_b
            d_W = self.d_W
            d_b += d_acts_BO.sum(axis=0)
            d_W += self.ops.batch_outer(d_acts_BO, acts_BI)
            if optimizer is not None and not kwargs.get('is_child'):
                optimizer(self.params.weights, self.params.gradient,
                    key=('', self.name), **kwargs)
            return self.ops.batch_dot(d_acts_BO, self.W.T)
        return finish_update



from cpython.mem cimport PyMem_Malloc, PyMem_Free, PyMem_Realloc
from libc.stdint cimport int32_t
from libc.string cimport memcpy
from ..typedefs cimport weight_t
from ..structs cimport SparseArrayC


cdef class SparseArray:
    cdef SparseArrayC* c

    @staticmethod
    cdef inline SparseArrayC* init(int key, weight_t value) except NULL:
        array = <SparseArrayC*>PyMem_Malloc(3 * sizeof(SparseArrayC))
        array[0] = SparseArrayC(key=key, val=value)
        array[1] = SparseArrayC(key=-1, val=0) # -1 marks end of values
        array[2] = SparseArrayC(key=-2, val=0) # -2 marks end of memory
        return array

    @staticmethod
    cdef inline SparseArrayC* clone(const SparseArrayC* src) except NULL:
        # Find length
        cdef int i = 0
        while src[i].key != -2:
            i += 1
        dest = <SparseArrayC*>PyMem_Malloc(i * sizeof(SparseArrayC))
        memcpy(dest, src, i * sizeof(SparseArrayC))
        return dest

    @staticmethod
    cdef inline int find_key(const SparseArrayC* array, int key) except -2:
        # Find either the key, or a place to insert it.
        cdef int i = 0
        while array[i].key != -2:
            if array[i].key == key:
                return i
            elif array[i].key == -1:
                return i
            else:
                i += 1
        else:
            return -1
    
    @staticmethod
    cdef inline SparseArrayC* resize(SparseArrayC* array) except NULL:
        cdef int length = 0
        while array[length].key != -2:
            length += 1
        cdef int new_length = length * 2
        array = <SparseArrayC*>PyMem_Realloc(array, new_length * sizeof(SparseArrayC))
        cdef int i
        for i in range(length, new_length-1):
            array[i] = SparseArrayC(key=-1, val=0)
        array[new_length-1] = SparseArrayC(key=-2, val=0)
        return array

    @staticmethod
    cdef inline int cmp(const void* a, const void* b) nogil:
        # Three-way comparison (I always forget):
        # 0 if equal
        # -1 if a before b
        # 1 if b before a 
        cdef int key_a = (<SparseArrayC*>a).key
        cdef int key_b = (<SparseArrayC*>b).key
        if key_a == key_b:
            return 0
        elif key_a < 0 or key_b < 0:
            return -1 if a > b else 1 # Sort negatives to the end 
        else:
            return -1 if a < b else 1



from .neural.vec2vec import Model


def layerize(begin_update=None, *args, **kwargs):
    '''Wrap a function into a layer'''
    if begin_update is not None:
        return FunctionLayer(begin_update, *args, **kwargs)
    def wrapper(begin_update):
        return FunctionLayer(begin_update, *args, **kwargs)
    return wrapper


def metalayerize(user_func):
    def returned(layers):
        forward, backward = split_backward(layers)
        def begin_update(X, *args, **kwargs):
            for func in forward:
                X = func(X)
            def finish_update(grad, *args, **kwargs):
                for bwd in backward:
                    grad = bwd(grad, *args, **kwargs)
                return grad
            return x, finish_update
        return FunctionLayer(begin_update, *args, **kwargs)
    return returned


def multiroute(output, activity, shapes, funcs):
    for i, (slice_, func) in enumerate(zip(shapes, funcs)):
        output[slice_] += func(output[slice_])
    return output


def sink_return(func, sink, splitter):
    def wrap(*args, **kwargs):
        output = func(*args, **kwargs)
        keep, sink = splitter(*output)
        sink(sink)
        return keep
    return wrap


def split_backward(layers):
    backward = []
    forward = [steal_callback(op.begin_update, backward.append)
               for op in layers]
    return forward, backward


class FunctionLayer(Model):
    def __init__(self, begin_update, predict_batch=None, predict_one=None,
            nr_in=None, nr_out=None, *args, **kwargs):
        self.begin_update = begin_update
        self.predict_batch = predict_batch
        self.predict_one = predict_one
        self.nr_in = nr_in
        self.nr_out = nr_out

    def __call__(self, X):
        if self.predict_batch is not None:
            return self.predict_batch(X)
        else:
            X, _ = self.begin_update(X)
            return X

    def check_input(self, X, expect_batch=False):
        return True
